import pytest
import numpy as np
from slicot import sg03bw


def test_sg03bw_basic_n1():
    """Test SG03BW with N=1 (basic case).

    Solves: A^T * X * C + E^T * X * D = SCALE * Y
    """
    trans = 'N'
    m, n = 2, 1

    # Simple upper triangular A and E
    a = np.array([[2.0, 1.0],
                  [0.0, 3.0]], dtype=np.float64, order='F')

    e = np.array([[1.0, 0.5],
                  [0.0, 1.0]], dtype=np.float64, order='F')

    c = np.array([[4.0]], dtype=np.float64, order='F')
    d = np.array([[2.0]], dtype=np.float64, order='F')

    # Right-hand side Y (will be overwritten with solution X)
    x = np.array([[1.0],
                  [2.0]], dtype=np.float64, order='F')

    x_out, scale, info = sg03bw(trans, a, e, c, d, x)

    # Check success
    assert info == 0, f"SG03BW failed with info={info}"

    # Scale should be positive
    assert 0 < scale <= 1.0, f"Invalid scale={scale}"

    # Solution should be finite
    assert np.all(np.isfinite(x_out))


def test_sg03bw_basic_n2():
    """Test SG03BW with N=2."""
    trans = 'N'
    m, n = 2, 2

    # Upper quasitriangular A (2x2 block)
    a = np.array([[1.0, 2.0],
                  [0.5, 3.0]], dtype=np.float64, order='F')

    # Upper triangular E
    e = np.array([[2.0, 1.0],
                  [0.0, 1.5]], dtype=np.float64, order='F')

    c = np.array([[1.0, 0.0],
                  [0.0, 1.0]], dtype=np.float64, order='F')

    d = np.array([[0.5, 0.0],
                  [0.0, 0.5]], dtype=np.float64, order='F')

    x = np.array([[1.0, 0.5],
                  [0.5, 1.0]], dtype=np.float64, order='F')

    x_out, scale, info = sg03bw(trans, a, e, c, d, x)

    assert info == 0, f"SG03BW failed with info={info}"
    assert 0 < scale <= 1.0, f"Invalid scale={scale}"
    assert np.all(np.isfinite(x_out))


def test_sg03bw_transposed():
    """Test SG03BW with TRANS='T' (transposed equation)."""
    trans = 'T'
    m, n = 2, 1

    a = np.array([[3.0, 1.0],
                  [0.0, 2.0]], dtype=np.float64, order='F')

    e = np.array([[1.5, 0.5],
                  [0.0, 1.0]], dtype=np.float64, order='F')

    c = np.array([[2.0]], dtype=np.float64, order='F')
    d = np.array([[1.0]], dtype=np.float64, order='F')

    x = np.array([[1.0],
                  [1.0]], dtype=np.float64, order='F')

    x_out, scale, info = sg03bw(trans, a, e, c, d, x)

    assert info == 0, f"SG03BW failed with info={info}"
    assert 0 < scale <= 1.0, f"Invalid scale={scale}"
    assert np.all(np.isfinite(x_out))


def test_sg03bw_m1_quick_return():
    """Test SG03BW with M=1 (minimal case for quick return)."""
    trans = 'N'
    m, n = 1, 1

    a = np.array([[1.0]], dtype=np.float64, order='F')
    e = np.array([[1.0]], dtype=np.float64, order='F')
    c = np.array([[1.0]], dtype=np.float64, order='F')
    d = np.array([[1.0]], dtype=np.float64, order='F')
    x = np.array([[0.0]], dtype=np.float64, order='F')

    x_out, scale, info = sg03bw(trans, a, e, c, d, x)

    # Should work without error
    assert info == 0, f"SG03BW failed with info={info}"
    assert scale > 0, f"Expected positive scale, got {scale}"


def test_sg03bw_invalid_trans():
    """Test SG03BW with invalid TRANS parameter."""
    trans = 'X'  # Invalid
    m, n = 2, 1

    a = np.eye(2, dtype=np.float64, order='F')
    e = np.eye(2, dtype=np.float64, order='F')
    c = np.ones((1, 1), dtype=np.float64, order='F')
    d = np.ones((1, 1), dtype=np.float64, order='F')
    x = np.ones((2, 1), dtype=np.float64, order='F')

    x_out, scale, info = sg03bw(trans, a, e, c, d, x)

    # Should return error
    assert info < 0, f"Expected info < 0 for invalid TRANS, got {info}"


def test_sg03bw_invalid_n():
    """Test SG03BW with invalid N (must be 1 or 2)."""
    trans = 'N'
    m, n = 2, 3  # Invalid N

    a = np.eye(2, dtype=np.float64, order='F')
    e = np.eye(2, dtype=np.float64, order='F')
    c = np.eye(3, dtype=np.float64, order='F')
    d = np.eye(3, dtype=np.float64, order='F')
    x = np.ones((2, 3), dtype=np.float64, order='F')

    x_out, scale, info = sg03bw(trans, a, e, c, d, x)

    # Should return error
    assert info < 0, f"Expected info < 0 for invalid N, got {info}"


def test_sg03bw_nearly_singular():
    """Test SG03BW with nearly singular system.

    Should return INFO=1 and use perturbed values.
    """
    trans = 'N'
    m, n = 2, 1

    # Create nearly singular system
    # A and E structured to make equation nearly singular
    a = np.array([[1e-15, 1.0],
                  [0.0, 1e-15]], dtype=np.float64, order='F')

    e = np.array([[1e-15, 0.5],
                  [0.0, 1e-15]], dtype=np.float64, order='F')

    c = np.array([[1.0]], dtype=np.float64, order='F')
    d = np.array([[1.0]], dtype=np.float64, order='F')

    x = np.array([[1.0],
                  [1.0]], dtype=np.float64, order='F')

    x_out, scale, info = sg03bw(trans, a, e, c, d, x)

    # May return INFO=1 for near singularity
    # But should still produce a finite result
    assert info >= 0, f"Unexpected error: info={info}"
    assert np.all(np.isfinite(x_out))
    assert 0 < scale <= 1.0


def test_sg03bw_identity_matrices():
    """Test SG03BW with identity matrices."""
    trans = 'N'
    m, n = 3, 1

    a = np.eye(3, dtype=np.float64, order='F')
    e = np.eye(3, dtype=np.float64, order='F')
    c = np.array([[2.0]], dtype=np.float64, order='F')
    d = np.array([[3.0]], dtype=np.float64, order='F')

    # Y = [1, 2, 3]^T
    x = np.array([[1.0],
                  [2.0],
                  [3.0]], dtype=np.float64, order='F')

    x_out, scale, info = sg03bw(trans, a, e, c, d, x)

    assert info == 0, f"SG03BW failed with info={info}"
    assert scale > 0, f"Invalid scale={scale}"

    # For identity matrices: (A^T + E^T) * X * (C + D) = (I + I) * X * (C + D)
    # = 2X * (C+D) = SCALE * Y
    # So X should be proportional to Y / (2*(C+D))
    expected_factor = scale / (2.0 * (c[0, 0] + d[0, 0]))
    expected_x = x * expected_factor / scale

    # Solution should have correct structure (may be scaled)
    assert np.all(np.isfinite(x_out))


def test_sg03bw_quasitriangular_boundary_m3():
    """Test SG03BW with M=3 to check boundary conditions for 2x2 blocks.

    This tests the fix for array bounds checking when detecting
    quasitriangular blocks at matrix boundaries.
    """
    trans = 'N'
    m, n = 3, 1

    # Upper quasitriangular with 2x2 block at positions [1:3, 1:3]
    # Block structure forces boundary checks
    a = np.array([[2.0, 1.0, 0.5],
                  [0.0, 1.0, 2.0],
                  [0.0, 0.5, 3.0]], dtype=np.float64, order='F')

    e = np.array([[1.0, 0.5, 0.0],
                  [0.0, 1.5, 1.0],
                  [0.0, 0.0, 2.0]], dtype=np.float64, order='F')

    c = np.array([[1.0]], dtype=np.float64, order='F')
    d = np.array([[1.0]], dtype=np.float64, order='F')

    x = np.array([[1.0],
                  [2.0],
                  [3.0]], dtype=np.float64, order='F')

    x_out, scale, info = sg03bw(trans, a, e, c, d, x)

    assert info == 0, f"SG03BW failed with info={info}"
    assert 0 < scale <= 1.0, f"Invalid scale={scale}"
    assert np.all(np.isfinite(x_out))


def test_sg03bw_quasitriangular_boundary_m4():
    """Test SG03BW with M=4 to verify 2x2 block handling at boundaries."""
    trans = 'T'  # Test transposed path
    m, n = 4, 1

    # Quasitriangular with multiple blocks
    a = np.array([[1.0, 2.0, 0.0, 0.0],
                  [0.5, 3.0, 1.0, 0.0],
                  [0.0, 0.0, 2.0, 1.5],
                  [0.0, 0.0, 0.5, 4.0]], dtype=np.float64, order='F')

    e = np.array([[2.0, 1.0, 0.0, 0.0],
                  [0.0, 1.5, 0.5, 0.0],
                  [0.0, 0.0, 1.0, 0.5],
                  [0.0, 0.0, 0.0, 2.0]], dtype=np.float64, order='F')

    c = np.array([[2.0]], dtype=np.float64, order='F')
    d = np.array([[1.5]], dtype=np.float64, order='F')

    x = np.array([[1.0],
                  [2.0],
                  [3.0],
                  [4.0]], dtype=np.float64, order='F')

    x_out, scale, info = sg03bw(trans, a, e, c, d, x)

    assert info == 0, f"SG03BW failed with info={info}"
    assert 0 < scale <= 1.0, f"Invalid scale={scale}"
    assert np.all(np.isfinite(x_out))



#!/usr/bin/env python3
"""
pytest configuration for SLICOT tests.

KNOWN ISSUE (macOS):
When running the full test suite (2600+ tests) on macOS, pytest may crash
with "Fatal Python error: Segmentation fault" or "Aborted" after ~750 tests. 
This is caused by a bug in macOS Accelerate framework (BLAS/LAPACK) that 
corrupts internal state or stack after many LAPACK calls.

Workaround: Run tests in batches.
A script is provided: `tools/run_tests_batched.py`
Usage: python3 tools/run_tests_batched.py

Note: Mixing OpenBLAS with Accelerate-linked NumPy causes immediate crashes,
so switching BLAS is not a viable solution unless NumPy is also rebuilt.

ASAN on Linux passes all tests - no memory issues detected.
"""
import ctypes
import os
import sys

import pytest

@pytest.fixture(autouse=True)
def flush_c_stderr():
    """Flush C stdio buffers after each test.

    This is a partial mitigation for macOS Accelerate issues - it flushes
    C-level stdio buffers after each test to reduce accumulation effects.
    Does not fully prevent the crash but may delay it.
    """
    yield
    if sys.platform == 'darwin':
        try:
            libc = ctypes.CDLL(None)
            libc.fflush(None)
        except Exception:
            pass


@pytest.fixture
def suppress_xerbla():
    """
    Context manager fixture to suppress XERBLA stderr output.

    XERBLA is a LAPACK error handler that prints to stderr when called
    with invalid parameters. This fixture redirects stderr to /dev/null
    during test execution to prevent delayed output appearing after
    test completion (especially on macOS).

    Usage:
        def test_invalid_param(suppress_xerbla):
            with suppress_xerbla():
                result = slicot_function('invalid', ...)
            assert result.info == -1
    """
    from contextlib import contextmanager

    @contextmanager
    def _suppress():
        stderr_fd = sys.stderr.fileno()
        saved_stderr = os.dup(stderr_fd)
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, stderr_fd)
        try:
            yield
        finally:
            # Flush C-level stderr before restoring
            try:
                libc = ctypes.CDLL(None)
                libc.fflush(None)
            except Exception:
                pass
            os.dup2(saved_stderr, stderr_fd)
            os.close(saved_stderr)
            os.close(devnull)

    return _suppress



"""
Tests for SB10JD - Convert descriptor state-space to regular state-space.

SB10JD converts the descriptor state-space system:
    E*dx/dt = A*x + B*u
         y = C*x + D*u
into regular state-space form:
    dx/dt = Ad*x + Bd*u
        y = Cd*x + Dd*u

Uses SVD decomposition for descriptor elimination.
"""
import numpy as np
import pytest
from slicot import sb10jd


"""Basic functionality tests."""

def test_identity_descriptor():
    """
    Test with E = I (identity). Output should match input.

    When E is identity, the descriptor system is already in regular form.
    """
    n, m, np_ = 3, 2, 2

    a = np.array([
        [-1.0, 0.5, 0.2],
        [0.0, -2.0, 0.3],
        [0.0, 0.0, -3.0]
    ], order='F', dtype=np.float64)

    b = np.array([
        [1.0, 0.0],
        [0.5, 1.0],
        [0.0, 0.5]
    ], order='F', dtype=np.float64)

    c = np.array([
        [1.0, 0.0, 0.0],
        [0.0, 1.0, 0.0]
    ], order='F', dtype=np.float64)

    d = np.array([
        [0.0, 0.0],
        [0.0, 0.0]
    ], order='F', dtype=np.float64)

    e = np.eye(n, order='F', dtype=np.float64)

    a_orig = a.copy()
    b_orig = b.copy()
    c_orig = c.copy()
    d_orig = d.copy()

    ad, bd, cd, dd, nsys, info = sb10jd(a, b, c, d, e)

    assert info == 0
    assert nsys == n
    np.testing.assert_allclose(ad[:nsys, :nsys], a_orig, rtol=1e-12)
    np.testing.assert_allclose(bd[:nsys, :], b_orig, rtol=1e-12)
    np.testing.assert_allclose(cd[:, :nsys], c_orig, rtol=1e-12)
    np.testing.assert_allclose(dd, d_orig, rtol=1e-12)

def test_scaled_identity_descriptor():
    """
    Test with E = 2*I (scaled identity).

    When E = 2*I, the regular form should preserve the transfer function.
    For E=2I, singular values are all 2, so transformation scales by 1/sqrt(2).
    Ad = (1/sqrt(2)) * A * (1/sqrt(2)) = A/2

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    n, m, np_ = 3, 2, 2

    a_orig = np.array([
        [-2.0, 1.0, 0.0],
        [0.0, -4.0, 1.0],
        [0.0, 0.0, -6.0]
    ], order='F', dtype=np.float64)

    b_orig = np.array([
        [2.0, 0.0],
        [0.0, 2.0],
        [1.0, 1.0]
    ], order='F', dtype=np.float64)

    c_orig = np.array([
        [1.0, 0.0, 1.0],
        [0.0, 1.0, 0.0]
    ], order='F', dtype=np.float64)

    d_orig = np.array([
        [0.1, 0.0],
        [0.0, 0.1]
    ], order='F', dtype=np.float64)

    e = 2.0 * np.eye(n, order='F', dtype=np.float64)

    a = a_orig.copy()
    b = b_orig.copy()
    c = c_orig.copy()
    d = d_orig.copy()

    ad, bd, cd, dd, nsys, info = sb10jd(a, b, c, d, e)

    assert info == 0
    assert nsys == n

    a_expected = a_orig / 2.0
    b_expected = b_orig / np.sqrt(2.0)
    c_expected = c_orig / np.sqrt(2.0)

    np.testing.assert_allclose(ad[:nsys, :nsys], a_expected, rtol=1e-10)
    np.testing.assert_allclose(bd[:nsys, :], b_expected, rtol=1e-10)
    np.testing.assert_allclose(cd[:, :nsys], c_expected, rtol=1e-10)


"""Tests for rank-deficient E matrices (descriptor elimination)."""

def test_rank_deficient_e_simple():
    """
    Test with rank-deficient E (nsys < n).

    E = [[1, 0], [0, 0]] has rank 1, so nsys should be 1.
    """
    n, m, np_ = 2, 1, 1

    a = np.array([
        [1.0, 2.0],
        [3.0, 4.0]
    ], order='F', dtype=np.float64)

    b = np.array([
        [1.0],
        [1.0]
    ], order='F', dtype=np.float64)

    c = np.array([
        [1.0, 0.0]
    ], order='F', dtype=np.float64)

    d = np.array([
        [0.0]
    ], order='F', dtype=np.float64)

    e = np.array([
        [1.0, 0.0],
        [0.0, 0.0]
    ], order='F', dtype=np.float64)

    ad, bd, cd, dd, nsys, info = sb10jd(a, b, c, d, e)

    assert info == 0
    assert nsys == 1

def test_rank_deficient_e_2x3():
    """
    Test rank-2 E with n=3.

    E matrix with rank 2 should produce nsys=2.
    """
    n, m, np_ = 3, 2, 2

    a = np.array([
        [-1.0, 1.0, 0.0],
        [0.0, -2.0, 1.0],
        [1.0, 0.0, -3.0]
    ], order='F', dtype=np.float64)

    b = np.array([
        [1.0, 0.0],
        [0.0, 1.0],
        [0.5, 0.5]
    ], order='F', dtype=np.float64)

    c = np.array([
        [1.0, 0.0, 0.0],
        [0.0, 1.0, 0.0]
    ], order='F', dtype=np.float64)

    d = np.zeros((np_, m), order='F', dtype=np.float64)

    e = np.array([
        [1.0, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        [0.0, 0.0, 0.0]
    ], order='F', dtype=np.float64)

    ad, bd, cd, dd, nsys, info = sb10jd(a, b, c, d, e)

    assert info == 0
    assert nsys == 2


"""Test transfer function preservation property."""

def test_transfer_function_preservation():
    """
    Verify transfer function is preserved.

    G(s) = C*(s*E - A)^(-1)*B + D should equal
    Gd(s) = Cd*(s*I - Ad)^(-1)*Bd + Dd at test frequencies.

    Random seed: 100 (for reproducibility)
    """
    np.random.seed(100)
    n, m, np_ = 3, 2, 2

    a = np.array([
        [-2.0, 0.5, 0.0],
        [0.0, -3.0, 0.5],
        [0.0, 0.0, -4.0]
    ], order='F', dtype=np.float64)

    b = np.array([
        [1.0, 0.0],
        [0.5, 1.0],
        [0.0, 0.5]
    ], order='F', dtype=np.float64)

    c = np.array([
        [1.0, 0.0, 0.5],
        [0.0, 1.0, 0.0]
    ], order='F', dtype=np.float64)

    d = np.array([
        [0.1, 0.0],
        [0.0, 0.1]
    ], order='F', dtype=np.float64)

    e = np.eye(n, order='F', dtype=np.float64)

    a_orig = a.copy()
    b_orig = b.copy()
    c_orig = c.copy()
    d_orig = d.copy()
    e_orig = e.copy()

    ad, bd, cd, dd, nsys, info = sb10jd(a, b, c, d, e)

    assert info == 0

    test_freqs = [0.1, 1.0, 10.0, 100.0]
    for omega in test_freqs:
        s = 1j * omega

        g_desc = c_orig @ np.linalg.solve(s * e_orig - a_orig, b_orig) + d_orig

        ad_sub = ad[:nsys, :nsys]
        bd_sub = bd[:nsys, :]
        cd_sub = cd[:, :nsys]
        g_regular = cd_sub @ np.linalg.solve(s * np.eye(nsys) - ad_sub, bd_sub) + dd

        np.testing.assert_allclose(g_regular, g_desc, rtol=1e-10)


"""Test with zero or near-zero E matrix."""

def test_zero_e_matrix():
    """
    Test with E = 0 (all singular values below tolerance).

    When E = 0, A becomes -1/eps*I on diagonal, B and C become zero.
    """
    n, m, np_ = 2, 1, 1

    a = np.array([
        [1.0, 2.0],
        [3.0, 4.0]
    ], order='F', dtype=np.float64)

    b = np.array([
        [1.0],
        [1.0]
    ], order='F', dtype=np.float64)

    c = np.array([
        [1.0, 0.0]
    ], order='F', dtype=np.float64)

    d = np.array([
        [0.5]
    ], order='F', dtype=np.float64)

    e = np.zeros((n, n), order='F', dtype=np.float64)

    ad, bd, cd, dd, nsys, info = sb10jd(a, b, c, d, e)

    assert info == 0
    assert nsys == n

    np.testing.assert_allclose(bd[:nsys, :], 0.0, atol=1e-10)
    np.testing.assert_allclose(cd[:, :nsys], 0.0, atol=1e-10)


"""Edge cases and boundary conditions."""

def test_n_zero():
    """Test N=0 quick return."""
    n, m, np_ = 0, 2, 2

    a = np.zeros((1, 1), order='F', dtype=np.float64)
    b = np.zeros((1, m), order='F', dtype=np.float64)
    c = np.zeros((np_, 1), order='F', dtype=np.float64)
    d = np.zeros((np_, m), order='F', dtype=np.float64)
    e = np.zeros((1, 1), order='F', dtype=np.float64)

    ad, bd, cd, dd, nsys, info = sb10jd(a, b, c, d, e, n=0, m=m, np=np_)

    assert info == 0
    assert nsys == 0

def test_m_zero():
    """Test M=0 (no inputs)."""
    n, m, np_ = 2, 0, 1

    a = np.array([
        [-1.0, 0.5],
        [0.0, -2.0]
    ], order='F', dtype=np.float64)

    b = np.zeros((n, 1), order='F', dtype=np.float64)
    c = np.array([[1.0, 0.0]], order='F', dtype=np.float64)
    d = np.zeros((np_, 1), order='F', dtype=np.float64)
    e = np.eye(n, order='F', dtype=np.float64)

    ad, bd, cd, dd, nsys, info = sb10jd(a, b, c, d, e, m=0)

    assert info == 0
    assert nsys == n

def test_np_zero():
    """Test NP=0 (no outputs)."""
    n, m, np_ = 2, 1, 0

    a = np.array([
        [-1.0, 0.5],
        [0.0, -2.0]
    ], order='F', dtype=np.float64)

    b = np.array([[1.0], [0.5]], order='F', dtype=np.float64)
    c = np.zeros((1, n), order='F', dtype=np.float64)
    d = np.zeros((1, m), order='F', dtype=np.float64)
    e = np.eye(n, order='F', dtype=np.float64)

    ad, bd, cd, dd, nsys, info = sb10jd(a, b, c, d, e, np=0)

    assert info == 0
    assert nsys == n


"""Test error handling."""

def test_invalid_lda():
    """Test SVD failure returns info=1 (for zero E matrix with small tolerance edge case)."""
    n, m, np_ = 3, 2, 2

    a = np.eye(n, order='F', dtype=np.float64)
    b = np.ones((n, m), order='F', dtype=np.float64)
    c = np.ones((np_, n), order='F', dtype=np.float64)
    d = np.zeros((np_, m), order='F', dtype=np.float64)
    e = np.eye(n, order='F', dtype=np.float64)

    ad, bd, cd, dd, nsys, info = sb10jd(a, b, c, d, e)

    assert info == 0
    assert nsys == n


"""Test larger systems for robustness."""

def test_5x5_full_rank():
    """
    Test 5x5 system with full rank E.

    Random seed: 200 (for reproducibility)
    """
    np.random.seed(200)
    n, m, np_ = 5, 3, 2

    a = -np.eye(n) + 0.1 * np.random.randn(n, n)
    a = np.asfortranarray(a)

    b = np.random.randn(n, m)
    b = np.asfortranarray(b)

    c = np.random.randn(np_, n)
    c = np.asfortranarray(c)

    d = 0.1 * np.random.randn(np_, m)
    d = np.asfortranarray(d)

    e = np.eye(n) + 0.01 * np.random.randn(n, n)
    e = np.asfortranarray(e)

    a_orig = a.copy()
    b_orig = b.copy()
    c_orig = c.copy()
    d_orig = d.copy()
    e_orig = e.copy()

    ad, bd, cd, dd, nsys, info = sb10jd(a, b, c, d, e)

    assert info == 0
    assert nsys == n

    s = 1j * 1.0
    g_desc = c_orig @ np.linalg.solve(s * e_orig - a_orig, b_orig) + d_orig
    ad_sub = ad[:nsys, :nsys]
    bd_sub = bd[:nsys, :]
    cd_sub = cd[:, :nsys]
    g_regular = cd_sub @ np.linalg.solve(s * np.eye(nsys) - ad_sub, bd_sub) + dd

    np.testing.assert_allclose(g_regular, g_desc, rtol=1e-9)

def test_6x6_rank_deficient():
    """
    Test 6x6 system with rank-4 E.

    Random seed: 300 (for reproducibility)
    """
    np.random.seed(300)
    n, m, np_ = 6, 2, 3

    a = np.random.randn(n, n)
    a = np.asfortranarray(a)

    b = np.random.randn(n, m)
    b = np.asfortranarray(b)

    c = np.random.randn(np_, n)
    c = np.asfortranarray(c)

    d = np.random.randn(np_, m)
    d = np.asfortranarray(d)

    u_e = np.linalg.qr(np.random.randn(n, n))[0]
    s_e = np.diag([2.0, 1.5, 1.0, 0.5, 0.0, 0.0])
    e = u_e @ s_e @ u_e.T
    e = np.asfortranarray(e)

    ad, bd, cd, dd, nsys, info = sb10jd(a, b, c, d, e)

    assert info == 0
    assert nsys == 4



"""Tests for TF01MD - discrete-time state-space output response simulation."""

import numpy as np
import pytest
from slicot import tf01md


"""Basic functionality tests from SLICOT HTML documentation."""

def test_html_doc_example():
    """
    Test case from SLICOT HTML documentation.

    3x2x2 system (N=3 states, M=2 inputs, P=2 outputs), NY=10 time steps.

    Data read column-wise from:
    READ ( NIN, FMT = * ) ( ( A(I,J), I = 1,N ), J = 1,N )
    """
    n, m, p, ny = 3, 2, 2, 10

    # A (3x3) - read column-wise: ((A(I,J), I=1,N), J=1,N)
    # Data: 0.0000 -0.0700 0.0150 / 1.0000 0.8000 -0.1500 / 0.0000 0.0000 0.5000
    a = np.array([
        [0.0000, 1.0000, 0.0000],
        [-0.0700, 0.8000, 0.0000],
        [0.0150, -0.1500, 0.5000]
    ], order='F', dtype=float)

    # B (3x2) - read column-wise: ((B(I,J), I=1,N), J=1,M)
    # Data: 0.0000 2.0000 1.0000 / -1.0000 -0.1000 1.0000
    b = np.array([
        [0.0000, -1.0000],
        [2.0000, -0.1000],
        [1.0000, 1.0000]
    ], order='F', dtype=float)

    # C (2x3) - read column-wise: ((C(I,J), I=1,P), J=1,N)
    # Data: 0.0000 1.0000 / 0.0000 0.0000 / 1.0000 0.0000
    c = np.array([
        [0.0000, 0.0000, 1.0000],
        [1.0000, 0.0000, 0.0000]
    ], order='F', dtype=float)

    # D (2x2) - read column-wise: ((D(I,J), I=1,P), J=1,M)
    # Data: 1.0000 0.5000 / 0.0000 0.5000
    d = np.array([
        [1.0000, 0.0000],
        [0.5000, 0.5000]
    ], order='F', dtype=float)

    # x0 (3,) - initial state
    # Data: 1.0000 1.0000 1.0000
    x0 = np.array([1.0000, 1.0000, 1.0000], dtype=float)

    # U (2x10) - input sequence, read column-wise: ((U(I,J), I=1,M), J=1,NY)
    # Data spread across lines, read as: col1, col2, ...
    # -0.6922 -1.4934 / 0.3081 -2.7726 / 2.0039 0.2614 / ...
    u = np.array([
        [-0.6922, 0.3081, 2.0039, -0.9942, -1.5734, 0.4118, -0.9344, 0.8988, -0.0701, 0.0],
        [-1.4934, -2.7726, 0.2614, 1.8957, 1.5639, -1.4893, 1.2506, 0.2951, -0.9160, 0.0]
    ], order='F', dtype=float)

    # Actually, the U data needs re-parsing. Let me read it more carefully:
    # The data file shows:
    # -0.6922 -1.4934  0.3081 -2.7726  2.0039
    #  0.2614 -0.9160 -0.6030  1.2556  0.2951
    # -1.5734  1.5639 -0.9942  1.8957  0.8988
    #  0.4118 -1.4893 -0.9344  1.2506 -0.0701
    # This is 4 lines x 5 values = 20 values for M=2, NY=10
    # Read as ((U(I,J), I=1,M), J=1,NY) = column-wise
    # So first 2 values are U(:,1), next 2 are U(:,2), etc.
    u_flat = np.array([
        -0.6922, -1.4934, 0.3081, -2.7726, 2.0039,
        0.2614, -0.9160, -0.6030, 1.2556, 0.2951,
        -1.5734, 1.5639, -0.9942, 1.8957, 0.8988,
        0.4118, -1.4893, -0.9344, 1.2506, -0.0701
    ], dtype=float)
    # Column-wise: U(:,1) = first 2 values, U(:,2) = next 2, etc.
    u = u_flat.reshape((m, ny), order='F')

    # Expected outputs from HTML doc (P=2, NY=10)
    y_expected = np.array([
        [0.3078, -1.5125, -1.2577, -0.2947, -0.5632, -1.0846, -1.2427, 1.8097, 0.6685, -0.0896],
        [-0.0928, 1.2611, 3.4002, -0.7060, 5.4532, 1.1846, 2.2286, -1.9534, -4.4965, 1.1654]
    ], order='F', dtype=float)

    y, x_final, info = tf01md(a, b, c, d, u, x0.copy())

    assert info == 0
    np.testing.assert_allclose(y, y_expected, rtol=1e-3, atol=1e-4)

def test_state_update_property():
    """
    Validate state-space equations hold: x(k+1) = A*x(k) + B*u(k).

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    n, m, p = 3, 2, 2
    ny = 5

    a = np.random.randn(n, n).astype(float, order='F')
    b = np.random.randn(n, m).astype(float, order='F')
    c = np.random.randn(p, n).astype(float, order='F')
    d = np.random.randn(p, m).astype(float, order='F')
    u = np.random.randn(m, ny).astype(float, order='F')
    x0 = np.random.randn(n).astype(float)

    y, x_final, info = tf01md(a, b, c, d, u, x0.copy())

    assert info == 0

    # Manual simulation to verify
    x = x0.copy()
    for k in range(ny):
        y_k = c @ x + d @ u[:, k]
        x_next = a @ x + b @ u[:, k]
        np.testing.assert_allclose(y[:, k], y_k, rtol=1e-14, atol=1e-15)
        x = x_next

    np.testing.assert_allclose(x_final, x, rtol=1e-14, atol=1e-15)

def test_output_equation_property():
    """
    Validate output equation: y(k) = C*x(k) + D*u(k).

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    n, m, p = 4, 2, 3
    ny = 8

    a = np.random.randn(n, n).astype(float, order='F')
    b = np.random.randn(n, m).astype(float, order='F')
    c = np.random.randn(p, n).astype(float, order='F')
    d = np.random.randn(p, m).astype(float, order='F')
    u = np.random.randn(m, ny).astype(float, order='F')
    x0 = np.random.randn(n).astype(float)

    y, x_final, info = tf01md(a, b, c, d, u, x0.copy())

    assert info == 0

    # Simulate and check each output
    x = x0.copy()
    for k in range(ny):
        y_expected = c @ x + d @ u[:, k]
        np.testing.assert_allclose(y[:, k], y_expected, rtol=1e-14, atol=1e-15)
        x = a @ x + b @ u[:, k]


"""Edge case tests."""

def test_zero_time_steps():
    """Test with NY=0 (no time steps)."""
    n, m, p = 2, 1, 1
    a = np.eye(n, order='F', dtype=float)
    b = np.ones((n, m), order='F', dtype=float)
    c = np.ones((p, n), order='F', dtype=float)
    d = np.zeros((p, m), order='F', dtype=float)
    u = np.empty((m, 0), order='F', dtype=float)
    x0 = np.ones(n, dtype=float)

    y, x_final, info = tf01md(a, b, c, d, u, x0.copy())

    assert info == 0
    assert y.shape == (p, 0)

def test_zero_states():
    """Test with N=0 (pure feedthrough, no dynamics)."""
    n, m, p = 0, 2, 2
    ny = 5

    a = np.empty((n, n), order='F', dtype=float)
    b = np.empty((n, m), order='F', dtype=float)
    c = np.empty((p, n), order='F', dtype=float)
    d = np.array([[1.0, 0.5], [0.0, 2.0]], order='F', dtype=float)
    u = np.random.randn(m, ny).astype(float, order='F')
    x0 = np.empty(n, dtype=float)

    np.random.seed(456)
    u = np.random.randn(m, ny).astype(float, order='F')

    y, x_final, info = tf01md(a, b, c, d, u, x0)

    assert info == 0
    # y = D*u for each time step
    y_expected = d @ u
    np.testing.assert_allclose(y, y_expected, rtol=1e-14)

def test_single_input_single_output():
    """Test SISO system.

    Random seed: 789 (for reproducibility)
    """
    np.random.seed(789)
    n, m, p = 2, 1, 1
    ny = 10

    a = np.array([[0.9, 0.1], [-0.1, 0.8]], order='F', dtype=float)
    b = np.array([[1.0], [0.0]], order='F', dtype=float)
    c = np.array([[1.0, 0.0]], order='F', dtype=float)
    d = np.array([[0.0]], order='F', dtype=float)
    u = np.random.randn(m, ny).astype(float, order='F')
    x0 = np.zeros(n, dtype=float)

    y, x_final, info = tf01md(a, b, c, d, u, x0.copy())

    assert info == 0
    assert y.shape == (p, ny)

    # Manual simulation
    x = x0.copy()
    for k in range(ny):
        y_k = c @ x + d @ u[:, k]
        np.testing.assert_allclose(y[:, k], y_k.flatten(), rtol=1e-14)
        x = a @ x + b @ u[:, k].reshape(-1)


"""Error handling tests."""

def test_dimension_mismatch_a():
    """Test error with wrong A dimensions."""
    n, m, p = 3, 2, 2
    ny = 5

    a = np.eye(2, order='F', dtype=float)  # Wrong: should be 3x3
    b = np.zeros((n, m), order='F', dtype=float)
    c = np.zeros((p, n), order='F', dtype=float)
    d = np.zeros((p, m), order='F', dtype=float)
    u = np.zeros((m, ny), order='F', dtype=float)
    x0 = np.zeros(n, dtype=float)

    with pytest.raises(ValueError):
        tf01md(a, b, c, d, u, x0)



"""
Tests for AB05MD - Cascade (series) inter-connection of two systems.

AB05MD computes the state-space model (A,B,C,D) for the cascaded
inter-connection of two systems G1 and G2, where output of G1 feeds input of G2.
"""

import numpy as np
import pytest
from slicot import ab05md


"""Basic functionality tests using SLICOT HTML doc example."""

def test_lower_block_diagonal_form():
    """
    Test UPLO='L' using SLICOT HTML doc example.

    System 1: N1=3, M1=2, P1=2
    System 2: N2=3, P2=2

    For UPLO='L':
    A = [A1,    0   ]    B = [ B1   ]
        [B2*C1, A2  ]        [B2*D1 ]

    C = [D2*C1, C2]      D = D2*D1
    """
    a1 = np.array([
        [1.0, 0.0, -1.0],
        [0.0, -1.0, 1.0],
        [1.0, 1.0, 2.0]
    ], order='F', dtype=float)

    b1 = np.array([
        [1.0, 2.0],
        [1.0, 0.0],
        [0.0, 1.0]
    ], order='F', dtype=float)

    c1 = np.array([
        [3.0, -2.0, 1.0],
        [0.0, 1.0, 0.0]
    ], order='F', dtype=float)

    d1 = np.array([
        [1.0, 0.0],
        [0.0, 1.0]
    ], order='F', dtype=float)

    a2 = np.array([
        [-3.0, 0.0, 0.0],
        [1.0, 0.0, 1.0],
        [0.0, -1.0, 2.0]
    ], order='F', dtype=float)

    b2 = np.array([
        [0.0, 1.0],
        [-1.0, 0.0],
        [0.0, 2.0]
    ], order='F', dtype=float)

    c2 = np.array([
        [1.0, 1.0, 0.0],
        [1.0, 1.0, -1.0]
    ], order='F', dtype=float)

    d2 = np.array([
        [1.0, 1.0],
        [0.0, 1.0]
    ], order='F', dtype=float)

    a, b, c, d, n, info = ab05md('L', 'N', a1, b1, c1, d1, a2, b2, c2, d2)

    assert info == 0
    assert n == 6

    a_expected = np.array([
        [1.0, 0.0, -1.0, 0.0, 0.0, 0.0],
        [0.0, -1.0, 1.0, 0.0, 0.0, 0.0],
        [1.0, 1.0, 2.0, 0.0, 0.0, 0.0],
        [0.0, 1.0, 0.0, -3.0, 0.0, 0.0],
        [-3.0, 2.0, -1.0, 1.0, 0.0, 1.0],
        [0.0, 2.0, 0.0, 0.0, -1.0, 2.0]
    ], order='F', dtype=float)

    b_expected = np.array([
        [1.0, 2.0],
        [1.0, 0.0],
        [0.0, 1.0],
        [0.0, 1.0],
        [-1.0, 0.0],
        [0.0, 2.0]
    ], order='F', dtype=float)

    c_expected = np.array([
        [3.0, -1.0, 1.0, 1.0, 1.0, 0.0],
        [0.0, 1.0, 0.0, 1.0, 1.0, -1.0]
    ], order='F', dtype=float)

    d_expected = np.array([
        [1.0, 1.0],
        [0.0, 1.0]
    ], order='F', dtype=float)

    np.testing.assert_allclose(a, a_expected, rtol=1e-14)
    np.testing.assert_allclose(b, b_expected, rtol=1e-14)
    np.testing.assert_allclose(c, c_expected, rtol=1e-14)
    np.testing.assert_allclose(d, d_expected, rtol=1e-14)

def test_upper_block_diagonal_form():
    """
    Test UPLO='U' using same input data.

    For UPLO='U':
    A = [A2,    B2*C1]    B = [B2*D1]
        [0,     A1   ]        [ B1  ]

    C = [C2, D2*C1]      D = D2*D1
    """
    a1 = np.array([
        [1.0, 0.0, -1.0],
        [0.0, -1.0, 1.0],
        [1.0, 1.0, 2.0]
    ], order='F', dtype=float)

    b1 = np.array([
        [1.0, 2.0],
        [1.0, 0.0],
        [0.0, 1.0]
    ], order='F', dtype=float)

    c1 = np.array([
        [3.0, -2.0, 1.0],
        [0.0, 1.0, 0.0]
    ], order='F', dtype=float)

    d1 = np.array([
        [1.0, 0.0],
        [0.0, 1.0]
    ], order='F', dtype=float)

    a2 = np.array([
        [-3.0, 0.0, 0.0],
        [1.0, 0.0, 1.0],
        [0.0, -1.0, 2.0]
    ], order='F', dtype=float)

    b2 = np.array([
        [0.0, 1.0],
        [-1.0, 0.0],
        [0.0, 2.0]
    ], order='F', dtype=float)

    c2 = np.array([
        [1.0, 1.0, 0.0],
        [1.0, 1.0, -1.0]
    ], order='F', dtype=float)

    d2 = np.array([
        [1.0, 1.0],
        [0.0, 1.0]
    ], order='F', dtype=float)

    a, b, c, d, n, info = ab05md('U', 'N', a1, b1, c1, d1, a2, b2, c2, d2)

    assert info == 0
    assert n == 6

    b2c1 = b2 @ c1
    b2d1 = b2 @ d1
    d2c1 = d2 @ c1
    d2d1 = d2 @ d1

    a_expected = np.zeros((6, 6), order='F', dtype=float)
    a_expected[:3, :3] = a2
    a_expected[:3, 3:] = b2c1
    a_expected[3:, 3:] = a1

    b_expected = np.zeros((6, 2), order='F', dtype=float)
    b_expected[:3, :] = b2d1
    b_expected[3:, :] = b1

    c_expected = np.zeros((2, 6), order='F', dtype=float)
    c_expected[:, :3] = c2
    c_expected[:, 3:] = d2c1

    d_expected = d2d1

    np.testing.assert_allclose(a, a_expected, rtol=1e-14)
    np.testing.assert_allclose(b, b_expected, rtol=1e-14)
    np.testing.assert_allclose(c, c_expected, rtol=1e-14)
    np.testing.assert_allclose(d, d_expected, rtol=1e-14)


"""Test mathematical properties of cascade connection."""

def test_cascade_state_space_equations():
    """
    Validate that cascaded system satisfies state-space equations.

    For cascade G2*G1:
    - y = G2(G1(u)) = (C2*C1)*x + D2*D1*u (steady state through)
    - Markov parameters: h(0) = D = D2*D1

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    n1, m1, p1 = 2, 1, 2
    n2, p2 = 3, 1

    a1 = np.random.randn(n1, n1).astype(float, order='F') * 0.5
    b1 = np.random.randn(n1, m1).astype(float, order='F')
    c1 = np.random.randn(p1, n1).astype(float, order='F')
    d1 = np.random.randn(p1, m1).astype(float, order='F')

    a2 = np.random.randn(n2, n2).astype(float, order='F') * 0.5
    b2 = np.random.randn(n2, p1).astype(float, order='F')
    c2 = np.random.randn(p2, n2).astype(float, order='F')
    d2 = np.random.randn(p2, p1).astype(float, order='F')

    a, b, c, d, n, info = ab05md('L', 'N', a1, b1, c1, d1, a2, b2, c2, d2)

    assert info == 0
    assert n == n1 + n2

    d_expected = d2 @ d1
    np.testing.assert_allclose(d, d_expected, rtol=1e-14)

    cb = c @ b
    c1b1 = c1 @ b1
    c2b2 = c2 @ b2
    h1_expected = d2 @ c1b1 + c2 @ (b2 @ d1)
    np.testing.assert_allclose(cb, h1_expected, rtol=1e-13)

def test_cascade_transfer_function_equivalence():
    """
    Verify cascade connection preserves transfer function composition.

    For discrete-time system, simulate:
    y1 = G1(u), y2 = G2(y1) should equal y = G_cascade(u)

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    n1, m1, p1 = 2, 1, 2
    n2, p2 = 2, 1

    a1 = np.diag([0.5, 0.3]).astype(float, order='F')
    b1 = np.array([[1.0], [0.5]], order='F', dtype=float)
    c1 = np.array([[1.0, 0.0], [0.0, 1.0]], order='F', dtype=float)
    d1 = np.array([[0.1], [0.0]], order='F', dtype=float)

    a2 = np.diag([0.4, 0.6]).astype(float, order='F')
    b2 = np.array([[1.0, 0.5], [0.0, 1.0]], order='F', dtype=float)
    c2 = np.array([[1.0, 1.0]], order='F', dtype=float)
    d2 = np.array([[0.5, 0.2]], order='F', dtype=float)

    a, b, c, d, n, info = ab05md('L', 'N', a1, b1, c1, d1, a2, b2, c2, d2)
    assert info == 0

    nsteps = 10
    u = np.ones((m1, nsteps), order='F', dtype=float)

    x1 = np.zeros((n1, 1), order='F', dtype=float)
    x2 = np.zeros((n2, 1), order='F', dtype=float)
    y_separate = np.zeros((p2, nsteps), order='F', dtype=float)

    for k in range(nsteps):
        uk = u[:, k:k+1]
        y1 = c1 @ x1 + d1 @ uk
        y2 = c2 @ x2 + d2 @ y1
        y_separate[:, k:k+1] = y2
        x1 = a1 @ x1 + b1 @ uk
        x2 = a2 @ x2 + b2 @ y1

    x_casc = np.zeros((n, 1), order='F', dtype=float)
    y_cascade = np.zeros((p2, nsteps), order='F', dtype=float)

    for k in range(nsteps):
        uk = u[:, k:k+1]
        y_cascade[:, k:k+1] = c @ x_casc + d @ uk
        x_casc = a @ x_casc + b @ uk

    np.testing.assert_allclose(y_cascade, y_separate, rtol=1e-14, atol=1e-15)


"""Edge cases and boundary conditions."""

def test_zero_state_first_system():
    """
    Test cascade with N1=0 (first system is static gain D1 only).

    Random seed: 456 (for reproducibility)
    """
    np.random.seed(456)
    m1, p1 = 2, 2
    n2, p2 = 3, 2

    a1 = np.zeros((0, 0), order='F', dtype=float)
    b1 = np.zeros((0, m1), order='F', dtype=float)
    c1 = np.zeros((p1, 0), order='F', dtype=float)
    d1 = np.array([[1.0, 0.5], [0.0, 1.0]], order='F', dtype=float)

    a2 = np.random.randn(n2, n2).astype(float, order='F') * 0.5
    b2 = np.random.randn(n2, p1).astype(float, order='F')
    c2 = np.random.randn(p2, n2).astype(float, order='F')
    d2 = np.random.randn(p2, p1).astype(float, order='F')

    a, b, c, d, n, info = ab05md('L', 'N', a1, b1, c1, d1, a2, b2, c2, d2)

    assert info == 0
    assert n == n2

    np.testing.assert_allclose(a, a2, rtol=1e-14)
    np.testing.assert_allclose(b, b2 @ d1, rtol=1e-14)
    np.testing.assert_allclose(c, c2, rtol=1e-14)
    np.testing.assert_allclose(d, d2 @ d1, rtol=1e-14)

def test_zero_state_second_system():
    """
    Test cascade with N2=0 (second system is static gain D2 only).

    Random seed: 789 (for reproducibility)
    """
    np.random.seed(789)
    n1, m1, p1 = 3, 2, 2
    p2 = 2

    a1 = np.random.randn(n1, n1).astype(float, order='F') * 0.5
    b1 = np.random.randn(n1, m1).astype(float, order='F')
    c1 = np.random.randn(p1, n1).astype(float, order='F')
    d1 = np.random.randn(p1, m1).astype(float, order='F')

    a2 = np.zeros((0, 0), order='F', dtype=float)
    b2 = np.zeros((0, p1), order='F', dtype=float)
    c2 = np.zeros((p2, 0), order='F', dtype=float)
    d2 = np.array([[1.0, 0.5], [0.0, 1.0]], order='F', dtype=float)

    a, b, c, d, n, info = ab05md('L', 'N', a1, b1, c1, d1, a2, b2, c2, d2)

    assert info == 0
    assert n == n1

    np.testing.assert_allclose(a, a1, rtol=1e-14)
    np.testing.assert_allclose(b, b1, rtol=1e-14)
    np.testing.assert_allclose(c, d2 @ c1, rtol=1e-14)
    np.testing.assert_allclose(d, d2 @ d1, rtol=1e-14)

def test_single_state_systems():
    """
    Test with minimal N1=1, N2=1 systems.

    Random seed: 999 (for reproducibility)
    """
    np.random.seed(999)
    n1, m1, p1 = 1, 1, 1
    n2, p2 = 1, 1

    a1 = np.array([[0.5]], order='F', dtype=float)
    b1 = np.array([[1.0]], order='F', dtype=float)
    c1 = np.array([[2.0]], order='F', dtype=float)
    d1 = np.array([[0.1]], order='F', dtype=float)

    a2 = np.array([[0.3]], order='F', dtype=float)
    b2 = np.array([[1.5]], order='F', dtype=float)
    c2 = np.array([[0.8]], order='F', dtype=float)
    d2 = np.array([[0.2]], order='F', dtype=float)

    a, b, c, d, n, info = ab05md('L', 'N', a1, b1, c1, d1, a2, b2, c2, d2)

    assert info == 0
    assert n == 2

    a_expected = np.array([
        [0.5, 0.0],
        [1.5 * 2.0, 0.3]
    ], order='F', dtype=float)

    b_expected = np.array([
        [1.0],
        [1.5 * 0.1]
    ], order='F', dtype=float)

    c_expected = np.array([
        [0.2 * 2.0, 0.8]
    ], order='F', dtype=float)

    d_expected = np.array([
        [0.2 * 0.1]
    ], order='F', dtype=float)

    np.testing.assert_allclose(a, a_expected, rtol=1e-14)
    np.testing.assert_allclose(b, b_expected, rtol=1e-14)
    np.testing.assert_allclose(c, c_expected, rtol=1e-14)
    np.testing.assert_allclose(d, d_expected, rtol=1e-14)


"""Test error handling and invalid input detection."""

def test_invalid_uplo():
    """Test invalid UPLO parameter."""
    a1 = np.array([[1.0]], order='F', dtype=float)
    b1 = np.array([[1.0]], order='F', dtype=float)
    c1 = np.array([[1.0]], order='F', dtype=float)
    d1 = np.array([[1.0]], order='F', dtype=float)
    a2 = np.array([[1.0]], order='F', dtype=float)
    b2 = np.array([[1.0]], order='F', dtype=float)
    c2 = np.array([[1.0]], order='F', dtype=float)
    d2 = np.array([[1.0]], order='F', dtype=float)

    with pytest.raises(ValueError, match="uplo"):
        ab05md('X', 'N', a1, b1, c1, d1, a2, b2, c2, d2)

def test_invalid_over():
    """Test invalid OVER parameter."""
    a1 = np.array([[1.0]], order='F', dtype=float)
    b1 = np.array([[1.0]], order='F', dtype=float)
    c1 = np.array([[1.0]], order='F', dtype=float)
    d1 = np.array([[1.0]], order='F', dtype=float)
    a2 = np.array([[1.0]], order='F', dtype=float)
    b2 = np.array([[1.0]], order='F', dtype=float)
    c2 = np.array([[1.0]], order='F', dtype=float)
    d2 = np.array([[1.0]], order='F', dtype=float)

    with pytest.raises(ValueError, match="over"):
        ab05md('L', 'X', a1, b1, c1, d1, a2, b2, c2, d2)

def test_dimension_mismatch_p1():
    """Test dimension mismatch: P1 (output of G1 != input of G2)."""
    a1 = np.array([[1.0]], order='F', dtype=float)
    b1 = np.array([[1.0]], order='F', dtype=float)
    c1 = np.array([[1.0], [2.0]], order='F', dtype=float)
    d1 = np.array([[1.0], [2.0]], order='F', dtype=float)
    a2 = np.array([[1.0]], order='F', dtype=float)
    b2 = np.array([[1.0, 2.0, 3.0]], order='F', dtype=float)
    c2 = np.array([[1.0]], order='F', dtype=float)
    d2 = np.array([[1.0, 2.0, 3.0]], order='F', dtype=float)

    with pytest.raises(ValueError, match="[pP]1"):
        ab05md('L', 'N', a1, b1, c1, d1, a2, b2, c2, d2)



"""
Tests for SB03OZ - Complex Lyapunov equation solver computing Cholesky factor.

Solves for X = op(U)^H * op(U) either the stable continuous-time Lyapunov equation:
    op(A)^H * X + X * op(A) = -scale^2 * op(B)^H * op(B)
or the convergent discrete-time Lyapunov equation:
    op(A)^H * X * op(A) - X = -scale^2 * op(B)^H * op(B)

where A is N-by-N complex, op(B) is M-by-N complex, U is upper triangular Cholesky factor.

Tests via ctypes since SB03OZ uses complex arrays.
"""
import ctypes
import numpy as np
import os
import glob
import pytest


def find_slicot_library():
    """Find the slicot shared library."""
    project_root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    build_dirs = [
        os.path.join(project_root, 'build', 'macos-arm64-debug', 'src'),
        os.path.join(project_root, 'build', 'linux-x64-debug', 'src'),
        os.path.join(project_root, 'build', 'linux-x64-debug-sanitizers', 'src'),
        os.path.join(project_root, 'build', 'macos-arm64-release', 'src'),
        os.path.join(project_root, 'build', 'linux-x64-release', 'src'),
    ]
    for bd in build_dirs:
        if os.path.exists(bd):
            libs = glob.glob(os.path.join(bd, 'libslicot.*'))
            if libs:
                return libs[0]
    pytest.skip("Could not find libslicot shared library")


@pytest.fixture(scope='module')
def lib():
    """Load the slicot library."""
    lib_path = find_slicot_library()
    slicot = ctypes.CDLL(lib_path)
    return slicot


@pytest.fixture(scope='module')
def sb03oz(lib):
    """Get the sb03oz function with proper signature."""
    func = lib.sb03oz
    func.argtypes = [
        ctypes.c_char_p,                  # dico
        ctypes.c_char_p,                  # fact
        ctypes.c_char_p,                  # trans
        ctypes.c_int,                     # n
        ctypes.c_int,                     # m
        ctypes.POINTER(ctypes.c_double),  # a (complex, 2*doubles per element)
        ctypes.c_int,                     # lda
        ctypes.POINTER(ctypes.c_double),  # q (complex)
        ctypes.c_int,                     # ldq
        ctypes.POINTER(ctypes.c_double),  # b (complex)
        ctypes.c_int,                     # ldb
        ctypes.POINTER(ctypes.c_double),  # scale
        ctypes.POINTER(ctypes.c_double),  # w (complex eigenvalues)
        ctypes.POINTER(ctypes.c_double),  # dwork
        ctypes.POINTER(ctypes.c_double),  # zwork (complex workspace)
        ctypes.c_int,                     # lzwork
        ctypes.POINTER(ctypes.c_int),     # info
    ]
    func.restype = None
    return func


def make_stable_continuous_complex(n, seed=42):
    """
    Create an n-by-n complex matrix with stable eigenvalues (negative real parts).

    Random seed: {seed} (for reproducibility)
    """
    np.random.seed(seed)
    u, _ = np.linalg.qr(np.random.randn(n, n) + 1j*np.random.randn(n, n))
    eig = -np.random.rand(n) - 0.5 + 1j * np.random.randn(n) * 0.3
    a = u @ np.diag(eig) @ u.conj().T
    return np.asfortranarray(a)


def make_stable_discrete_complex(n, seed=42):
    """
    Create an n-by-n complex matrix with convergent eigenvalues (modulus < 1).

    Random seed: {seed} (for reproducibility)
    """
    np.random.seed(seed)
    u, _ = np.linalg.qr(np.random.randn(n, n) + 1j*np.random.randn(n, n))
    r = 0.3 + np.random.rand(n) * 0.5
    theta = np.random.rand(n) * 2 * np.pi
    eig = r * np.exp(1j * theta)
    a = u @ np.diag(eig) @ u.conj().T
    return np.asfortranarray(a)


def test_continuous_nofact_notrans(sb03oz):
    """
    Test continuous-time, compute Schur factorization, no transpose.

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    n, m = 3, 4

    a = make_stable_continuous_complex(n, seed=42)
    a_orig = a.copy()

    b = np.random.randn(m, n) + 1j * np.random.randn(m, n)
    b = np.asfortranarray(b)
    b_orig = b.copy()

    q = np.zeros((n, n), order='F', dtype=np.complex128)

    scale = ctypes.c_double(0.0)
    w = np.zeros(n, dtype=np.complex128)
    ldwork = max(1, n)
    dwork = np.zeros(ldwork, dtype=np.float64)
    lzwork = max(1, 2*n + max(min(n, m) - 2, 0))
    zwork = np.zeros(lzwork, dtype=np.complex128)
    info = ctypes.c_int(0)

    ldb = max(n, m)
    b_padded = np.zeros((ldb, max(m, n)), order='F', dtype=np.complex128)
    b_padded[:m, :n] = b

    sb03oz(
        b"C", b"N", b"N", n, m,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        q.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b_padded.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldb,
        ctypes.byref(scale),
        w.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        zwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), lzwork,
        ctypes.byref(info)
    )

    assert info.value == 0
    assert 0 < scale.value <= 1.0

    u = np.triu(b_padded[:n, :n])
    x = u.conj().T @ u

    rhs = -scale.value**2 * b_orig.conj().T @ b_orig
    residual = a_orig.conj().T @ x + x @ a_orig - rhs
    np.testing.assert_allclose(residual, 0.0, atol=1e-9)


def test_continuous_transpose(sb03oz):
    """
    Test continuous-time with transpose (TRANS='C').

    Equation: A * X + X * A^H = -scale^2 * B * B^H, X = U * U^H
    Random seed: 100 (for reproducibility)
    """
    np.random.seed(100)
    n, m = 3, 4

    a = make_stable_continuous_complex(n, seed=100)
    a_orig = a.copy()

    b = np.random.randn(n, m) + 1j * np.random.randn(n, m)
    b = np.asfortranarray(b)
    b_orig = b.copy()

    q = np.zeros((n, n), order='F', dtype=np.complex128)

    scale = ctypes.c_double(0.0)
    w = np.zeros(n, dtype=np.complex128)
    ldwork = max(1, n)
    dwork = np.zeros(ldwork, dtype=np.float64)
    lzwork = max(1, 2*n + max(min(n, m) - 2, 0))
    zwork = np.zeros(lzwork, dtype=np.complex128)
    info = ctypes.c_int(0)

    ldb = n
    b_padded = np.zeros((ldb, max(m, n)), order='F', dtype=np.complex128)
    b_padded[:n, :m] = b

    sb03oz(
        b"C", b"N", b"C", n, m,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        q.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b_padded.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldb,
        ctypes.byref(scale),
        w.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        zwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), lzwork,
        ctypes.byref(info)
    )

    assert info.value == 0
    assert 0 < scale.value <= 1.0

    u = np.triu(b_padded[:n, :n])
    x = u @ u.conj().T

    rhs = -scale.value**2 * b_orig @ b_orig.conj().T
    residual = a_orig @ x + x @ a_orig.conj().T - rhs
    np.testing.assert_allclose(residual, 0.0, atol=1e-9)


def test_discrete_nofact_notrans(sb03oz):
    """
    Test discrete-time, compute Schur factorization, no transpose.

    Equation: A^H * X * A - X = -scale^2 * B^H * B, X = U^H * U
    Random seed: 200 (for reproducibility)
    """
    np.random.seed(200)
    n, m = 3, 4

    a = make_stable_discrete_complex(n, seed=200)
    a_orig = a.copy()

    b = np.random.randn(m, n) + 1j * np.random.randn(m, n)
    b = np.asfortranarray(b)
    b_orig = b.copy()

    q = np.zeros((n, n), order='F', dtype=np.complex128)

    scale = ctypes.c_double(0.0)
    w = np.zeros(n, dtype=np.complex128)
    ldwork = max(1, n)
    dwork = np.zeros(ldwork, dtype=np.float64)
    lzwork = max(1, 2*n + max(min(n, m) - 2, 0))
    zwork = np.zeros(lzwork, dtype=np.complex128)
    info = ctypes.c_int(0)

    ldb = max(n, m)
    b_padded = np.zeros((ldb, max(m, n)), order='F', dtype=np.complex128)
    b_padded[:m, :n] = b

    sb03oz(
        b"D", b"N", b"N", n, m,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        q.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b_padded.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldb,
        ctypes.byref(scale),
        w.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        zwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), lzwork,
        ctypes.byref(info)
    )

    assert info.value == 0
    assert 0 < scale.value <= 1.0

    u = np.triu(b_padded[:n, :n])
    x = u.conj().T @ u

    rhs = -scale.value**2 * b_orig.conj().T @ b_orig
    residual = a_orig.conj().T @ x @ a_orig - x - rhs
    np.testing.assert_allclose(residual, 0.0, atol=1e-9)


def test_discrete_transpose(sb03oz):
    """
    Test discrete-time with transpose (TRANS='C').

    Equation: A * X * A^H - X = -scale^2 * B * B^H, X = U * U^H
    Random seed: 300 (for reproducibility)
    """
    np.random.seed(300)
    n, m = 3, 4

    a = make_stable_discrete_complex(n, seed=300)
    a_orig = a.copy()

    b = np.random.randn(n, m) + 1j * np.random.randn(n, m)
    b = np.asfortranarray(b)
    b_orig = b.copy()

    q = np.zeros((n, n), order='F', dtype=np.complex128)

    scale = ctypes.c_double(0.0)
    w = np.zeros(n, dtype=np.complex128)
    ldwork = max(1, n)
    dwork = np.zeros(ldwork, dtype=np.float64)
    lzwork = max(1, 2*n + max(min(n, m) - 2, 0))
    zwork = np.zeros(lzwork, dtype=np.complex128)
    info = ctypes.c_int(0)

    ldb = n
    b_padded = np.zeros((ldb, max(m, n)), order='F', dtype=np.complex128)
    b_padded[:n, :m] = b

    sb03oz(
        b"D", b"N", b"C", n, m,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        q.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b_padded.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldb,
        ctypes.byref(scale),
        w.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        zwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), lzwork,
        ctypes.byref(info)
    )

    assert info.value == 0
    assert 0 < scale.value <= 1.0

    u = np.triu(b_padded[:n, :n])
    x = u @ u.conj().T

    rhs = -scale.value**2 * b_orig @ b_orig.conj().T
    residual = a_orig @ x @ a_orig.conj().T - x - rhs
    np.testing.assert_allclose(residual, 0.0, atol=1e-9)


def test_schur_provided(sb03oz):
    """
    Test with Schur factorization already provided (FACT='F').

    Random seed: 400 (for reproducibility)
    """
    np.random.seed(400)
    n, m = 3, 4

    s = np.array([
        [-1.0 + 0.2j, 0.3 - 0.1j, 0.2 + 0.1j],
        [0.0 + 0j, -1.5 + 0.3j, 0.4 - 0.2j],
        [0.0 + 0j, 0.0 + 0j, -2.0 - 0.5j]
    ], order='F', dtype=np.complex128)

    q = np.eye(n, order='F', dtype=np.complex128)

    b = np.random.randn(m, n) + 1j * np.random.randn(m, n)
    b = np.asfortranarray(b)
    b_orig = b.copy()

    a_orig = q @ s @ q.conj().T

    scale = ctypes.c_double(0.0)
    w = np.zeros(n, dtype=np.complex128)
    ldwork = max(1, n)
    dwork = np.zeros(ldwork, dtype=np.float64)
    lzwork = max(1, 2*n + max(min(n, m) - 2, 0))
    zwork = np.zeros(lzwork, dtype=np.complex128)
    info = ctypes.c_int(0)

    ldb = max(n, m)
    b_padded = np.zeros((ldb, max(m, n)), order='F', dtype=np.complex128)
    b_padded[:m, :n] = b

    sb03oz(
        b"C", b"F", b"N", n, m,
        s.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        q.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b_padded.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldb,
        ctypes.byref(scale),
        w.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        zwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), lzwork,
        ctypes.byref(info)
    )

    assert info.value == 0
    assert 0 < scale.value <= 1.0

    u = np.triu(b_padded[:n, :n])
    x = u.conj().T @ u

    rhs = -scale.value**2 * b_orig.conj().T @ b_orig
    residual = a_orig.conj().T @ x + x @ a_orig - rhs
    np.testing.assert_allclose(residual, 0.0, atol=1e-9)


def test_m_zero(sb03oz):
    """Test M=0: U should be set to zero."""
    n, m = 3, 0

    a = np.array([
        [-1.0 + 0.2j, 0.3 - 0.1j, 0.2 + 0.1j],
        [0.0 + 0j, -1.5 + 0.3j, 0.4 - 0.2j],
        [0.0 + 0j, 0.0 + 0j, -2.0 - 0.5j]
    ], order='F', dtype=np.complex128)

    q = np.zeros((n, n), order='F', dtype=np.complex128)

    b = np.zeros((n, n), order='F', dtype=np.complex128)
    b[0, 0] = 999.0 + 0j

    scale = ctypes.c_double(0.0)
    w = np.zeros(n, dtype=np.complex128)
    dwork = np.zeros(n, dtype=np.float64)
    zwork = np.zeros(1, dtype=np.complex128)
    info = ctypes.c_int(0)

    sb03oz(
        b"C", b"N", b"N", n, m,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        q.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        ctypes.byref(scale),
        w.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        zwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), 1,
        ctypes.byref(info)
    )

    assert info.value == 0
    np.testing.assert_allclose(b, 0.0, atol=1e-14)


def test_n_zero(sb03oz):
    """Test N=0: quick return."""
    n, m = 0, 3

    a = np.zeros((1, 1), order='F', dtype=np.complex128)
    q = np.zeros((1, 1), order='F', dtype=np.complex128)
    b = np.zeros((m, 1), order='F', dtype=np.complex128)

    scale = ctypes.c_double(0.0)
    w = np.zeros(1, dtype=np.complex128)
    dwork = np.zeros(1, dtype=np.float64)
    zwork = np.zeros(1, dtype=np.complex128)
    info = ctypes.c_int(0)

    sb03oz(
        b"C", b"N", b"N", n, m,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), 1,
        q.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), 1,
        b.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), m,
        ctypes.byref(scale),
        w.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        zwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), 1,
        ctypes.byref(info)
    )

    assert info.value == 0


def test_unstable_continuous(sb03oz):
    """Test unstable A (non-negative real eigenvalue) returns info=2."""
    n, m = 2, 2

    a = np.array([
        [1.0 + 0.2j, 0.0 + 0j],
        [0.0 + 0j, -1.0 + 0.3j]
    ], order='F', dtype=np.complex128)

    q = np.zeros((n, n), order='F', dtype=np.complex128)

    b = np.array([
        [1.0 + 0j, 0.0 + 0j],
        [0.0 + 0j, 1.0 + 0j]
    ], order='F', dtype=np.complex128)

    scale = ctypes.c_double(0.0)
    w = np.zeros(n, dtype=np.complex128)
    ldwork = max(1, n)
    dwork = np.zeros(ldwork, dtype=np.float64)
    lzwork = max(1, 2*n + max(min(n, m) - 2, 0))
    zwork = np.zeros(lzwork, dtype=np.complex128)
    info = ctypes.c_int(0)

    sb03oz(
        b"C", b"N", b"N", n, m,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        q.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        ctypes.byref(scale),
        w.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        zwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), lzwork,
        ctypes.byref(info)
    )

    assert info.value == 2


def test_non_convergent_discrete(sb03oz):
    """Test non-convergent A (eigenvalue modulus >= 1) returns info=2."""
    n, m = 2, 2

    a = np.array([
        [1.5 + 0.3j, 0.0 + 0j],
        [0.0 + 0j, 0.5 + 0.2j]
    ], order='F', dtype=np.complex128)

    q = np.zeros((n, n), order='F', dtype=np.complex128)

    b = np.array([
        [1.0 + 0j, 0.0 + 0j],
        [0.0 + 0j, 1.0 + 0j]
    ], order='F', dtype=np.complex128)

    scale = ctypes.c_double(0.0)
    w = np.zeros(n, dtype=np.complex128)
    ldwork = max(1, n)
    dwork = np.zeros(ldwork, dtype=np.float64)
    lzwork = max(1, 2*n + max(min(n, m) - 2, 0))
    zwork = np.zeros(lzwork, dtype=np.complex128)
    info = ctypes.c_int(0)

    sb03oz(
        b"D", b"N", b"N", n, m,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        q.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        ctypes.byref(scale),
        w.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        zwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), lzwork,
        ctypes.byref(info)
    )

    assert info.value == 2


def test_invalid_dico(sb03oz):
    """Test invalid DICO parameter."""
    n, m = 2, 2

    a = np.array([[-1.0 + 0j, 0.0 + 0j], [0.0 + 0j, -1.0 + 0j]], order='F', dtype=np.complex128)
    q = np.zeros((n, n), order='F', dtype=np.complex128)
    b = np.array([[1.0 + 0j, 0.0 + 0j], [0.0 + 0j, 1.0 + 0j]], order='F', dtype=np.complex128)

    scale = ctypes.c_double(0.0)
    w = np.zeros(n, dtype=np.complex128)
    ldwork = max(1, n)
    dwork = np.zeros(ldwork, dtype=np.float64)
    lzwork = max(1, 2*n + max(min(n, m) - 2, 0))
    zwork = np.zeros(lzwork, dtype=np.complex128)
    info = ctypes.c_int(0)

    sb03oz(
        b"X", b"N", b"N", n, m,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        q.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        ctypes.byref(scale),
        w.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        zwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), lzwork,
        ctypes.byref(info)
    )

    assert info.value == -1


def test_negative_n(sb03oz):
    """Test negative N."""
    n, m = -1, 2

    a = np.zeros((1, 1), order='F', dtype=np.complex128)
    q = np.zeros((1, 1), order='F', dtype=np.complex128)
    b = np.zeros((m, 1), order='F', dtype=np.complex128)

    scale = ctypes.c_double(0.0)
    w = np.zeros(1, dtype=np.complex128)
    dwork = np.zeros(4, dtype=np.float64)
    zwork = np.zeros(4, dtype=np.complex128)
    info = ctypes.c_int(0)

    sb03oz(
        b"C", b"N", b"N", n, m,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), 1,
        q.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), 1,
        b.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), 1,
        ctypes.byref(scale),
        w.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        zwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), 4,
        ctypes.byref(info)
    )

    assert info.value == -4


def test_5x5_continuous(sb03oz):
    """
    Test larger 5x5 continuous-time.

    Random seed: 500 (for reproducibility)
    """
    np.random.seed(500)
    n, m = 5, 7

    a = make_stable_continuous_complex(n, seed=500)
    a_orig = a.copy()

    b = np.random.randn(m, n) + 1j * np.random.randn(m, n)
    b = np.asfortranarray(b)
    b_orig = b.copy()

    q = np.zeros((n, n), order='F', dtype=np.complex128)

    scale = ctypes.c_double(0.0)
    w = np.zeros(n, dtype=np.complex128)
    ldwork = max(1, n)
    dwork = np.zeros(ldwork, dtype=np.float64)
    lzwork = max(1, 2*n + max(min(n, m) - 2, 0))
    zwork = np.zeros(lzwork, dtype=np.complex128)
    info = ctypes.c_int(0)

    ldb = max(n, m)
    b_padded = np.zeros((ldb, max(m, n)), order='F', dtype=np.complex128)
    b_padded[:m, :n] = b

    sb03oz(
        b"C", b"N", b"N", n, m,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        q.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b_padded.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldb,
        ctypes.byref(scale),
        w.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        zwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), lzwork,
        ctypes.byref(info)
    )

    assert info.value == 0
    assert 0 < scale.value <= 1.0

    u = np.triu(b_padded[:n, :n])
    x = u.conj().T @ u

    rhs = -scale.value**2 * b_orig.conj().T @ b_orig
    residual = a_orig.conj().T @ x + x @ a_orig - rhs
    np.testing.assert_allclose(residual, 0.0, atol=1e-8)


def test_positive_semidefinite(sb03oz):
    """
    Validate X = U^H * U is positive semi-definite (Hermitian positive).

    Random seed: 600 (for reproducibility)
    """
    np.random.seed(600)
    n, m = 4, 5

    a = make_stable_continuous_complex(n, seed=600)

    b = np.random.randn(m, n) + 1j * np.random.randn(m, n)
    b = np.asfortranarray(b)

    q = np.zeros((n, n), order='F', dtype=np.complex128)

    scale = ctypes.c_double(0.0)
    w = np.zeros(n, dtype=np.complex128)
    ldwork = max(1, n)
    dwork = np.zeros(ldwork, dtype=np.float64)
    lzwork = max(1, 2*n + max(min(n, m) - 2, 0))
    zwork = np.zeros(lzwork, dtype=np.complex128)
    info = ctypes.c_int(0)

    ldb = max(n, m)
    b_padded = np.zeros((ldb, max(m, n)), order='F', dtype=np.complex128)
    b_padded[:m, :n] = b

    sb03oz(
        b"C", b"N", b"N", n, m,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        q.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b_padded.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldb,
        ctypes.byref(scale),
        w.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        zwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), lzwork,
        ctypes.byref(info)
    )

    if info.value == 0:
        u = np.triu(b_padded[:n, :n])
        x = u.conj().T @ u
        eig = np.linalg.eigvalsh(x)
        assert all(e >= -1e-10 for e in eig), "X not positive semi-definite"


def test_upper_triangular_output(sb03oz):
    """
    Validate output U remains upper triangular with real non-negative diagonal.

    Random seed: 700 (for reproducibility)
    """
    np.random.seed(700)
    n, m = 4, 5

    a = make_stable_continuous_complex(n, seed=700)

    b = np.random.randn(m, n) + 1j * np.random.randn(m, n)
    b = np.asfortranarray(b)

    q = np.zeros((n, n), order='F', dtype=np.complex128)

    scale = ctypes.c_double(0.0)
    w = np.zeros(n, dtype=np.complex128)
    ldwork = max(1, n)
    dwork = np.zeros(ldwork, dtype=np.float64)
    lzwork = max(1, 2*n + max(min(n, m) - 2, 0))
    zwork = np.zeros(lzwork, dtype=np.complex128)
    info = ctypes.c_int(0)

    ldb = max(n, m)
    b_padded = np.zeros((ldb, max(m, n)), order='F', dtype=np.complex128)
    b_padded[:m, :n] = b

    sb03oz(
        b"C", b"N", b"N", n, m,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        q.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b_padded.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldb,
        ctypes.byref(scale),
        w.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        zwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), lzwork,
        ctypes.byref(info)
    )

    if info.value == 0:
        u = b_padded[:n, :n]
        assert np.allclose(np.tril(u, -1), 0, atol=1e-14), "U not upper triangular"
        for i in range(n):
            diag_val = u[i, i]
            assert abs(diag_val.imag) < 1e-14, f"Diagonal {i} not real"
            assert diag_val.real >= -1e-14, f"Diagonal {i} negative"


def test_eigenvalue_check(sb03oz):
    """
    Validate eigenvalues W returned match eigenvalues of A.

    Random seed: 800 (for reproducibility)
    """
    np.random.seed(800)
    n, m = 4, 5

    a = make_stable_continuous_complex(n, seed=800)
    a_orig = a.copy()

    b = np.random.randn(m, n) + 1j * np.random.randn(m, n)
    b = np.asfortranarray(b)

    q = np.zeros((n, n), order='F', dtype=np.complex128)

    scale = ctypes.c_double(0.0)
    w = np.zeros(n, dtype=np.complex128)
    ldwork = max(1, n)
    dwork = np.zeros(ldwork, dtype=np.float64)
    lzwork = max(1, 2*n + max(min(n, m) - 2, 0))
    zwork = np.zeros(lzwork, dtype=np.complex128)
    info = ctypes.c_int(0)

    ldb = max(n, m)
    b_padded = np.zeros((ldb, max(m, n)), order='F', dtype=np.complex128)
    b_padded[:m, :n] = b

    sb03oz(
        b"C", b"N", b"N", n, m,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        q.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b_padded.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldb,
        ctypes.byref(scale),
        w.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
        zwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), lzwork,
        ctypes.byref(info)
    )

    if info.value == 0:
        expected_eig = np.linalg.eigvals(a_orig)
        np.testing.assert_allclose(sorted(w.real), sorted(expected_eig.real), rtol=1e-10)
        np.testing.assert_allclose(sorted(w.imag), sorted(expected_eig.imag), rtol=1e-10)



"""
Tests for AB05OD - Rowwise concatenation of two systems in state-space form.

AB05OD computes the state-space model (A,B,C,D) for rowwise concatenation
(parallel inter-connection on outputs, with separate inputs) of two systems:
    Y = G1*U1 + alpha*G2*U2

The interconnected system has:
- A = [[A1, 0], [0, A2]]  (block diagonal)
- B = [[B1, 0], [0, B2]]  (block diagonal)
- C = [C1, alpha*C2]      (rowwise concatenation)
- D = [D1, alpha*D2]      (rowwise concatenation)
"""

import numpy as np
import pytest
from slicot import ab05od


"""Basic functionality tests using SLICOT HTML doc example."""

def test_html_example_alpha_one():
    """
    Test rowwise concatenation using SLICOT HTML doc example.

    System 1: N1=3, M1=2, P1=2
    System 2: N2=3, M2=2
    ALPHA=1.0

    From AB05OD.html Program Data and Results.
    """
    a1 = np.array([
        [1.0, 0.0, -1.0],
        [0.0, -1.0, 1.0],
        [1.0, 1.0, 2.0]
    ], order='F', dtype=float)

    b1 = np.array([
        [1.0, 2.0],
        [1.0, 0.0],
        [0.0, 1.0]
    ], order='F', dtype=float)

    c1 = np.array([
        [3.0, -2.0, 1.0],
        [0.0, 1.0, 0.0]
    ], order='F', dtype=float)

    d1 = np.array([
        [1.0, 0.0],
        [0.0, 1.0]
    ], order='F', dtype=float)

    a2 = np.array([
        [-3.0, 0.0, 0.0],
        [1.0, 0.0, 1.0],
        [0.0, -1.0, 2.0]
    ], order='F', dtype=float)

    b2 = np.array([
        [0.0, 1.0],
        [-1.0, 0.0],
        [0.0, 2.0]
    ], order='F', dtype=float)

    c2 = np.array([
        [1.0, 1.0, 0.0],
        [1.0, 1.0, -1.0]
    ], order='F', dtype=float)

    d2 = np.array([
        [1.0, 1.0],
        [0.0, 1.0]
    ], order='F', dtype=float)

    alpha = 1.0

    a, b, c, d, n, m, info = ab05od('N', a1, b1, c1, d1, a2, b2, c2, d2, alpha)

    assert info == 0
    assert n == 6
    assert m == 4

    a_expected = np.array([
        [1.0000, 0.0000, -1.0000, 0.0000, 0.0000, 0.0000],
        [0.0000, -1.0000, 1.0000, 0.0000, 0.0000, 0.0000],
        [1.0000, 1.0000, 2.0000, 0.0000, 0.0000, 0.0000],
        [0.0000, 0.0000, 0.0000, -3.0000, 0.0000, 0.0000],
        [0.0000, 0.0000, 0.0000, 1.0000, 0.0000, 1.0000],
        [0.0000, 0.0000, 0.0000, 0.0000, -1.0000, 2.0000]
    ], order='F', dtype=float)

    b_expected = np.array([
        [1.0000, 2.0000, 0.0000, 0.0000],
        [1.0000, 0.0000, 0.0000, 0.0000],
        [0.0000, 1.0000, 0.0000, 0.0000],
        [0.0000, 0.0000, 0.0000, 1.0000],
        [0.0000, 0.0000, -1.0000, 0.0000],
        [0.0000, 0.0000, 0.0000, 2.0000]
    ], order='F', dtype=float)

    c_expected = np.array([
        [3.0000, -2.0000, 1.0000, 1.0000, 1.0000, 0.0000],
        [0.0000, 1.0000, 0.0000, 1.0000, 1.0000, -1.0000]
    ], order='F', dtype=float)

    d_expected = np.array([
        [1.0000, 0.0000, 1.0000, 1.0000],
        [0.0000, 1.0000, 0.0000, 1.0000]
    ], order='F', dtype=float)

    np.testing.assert_allclose(a, a_expected, rtol=1e-3, atol=1e-4)
    np.testing.assert_allclose(b, b_expected, rtol=1e-3, atol=1e-4)
    np.testing.assert_allclose(c, c_expected, rtol=1e-3, atol=1e-4)
    np.testing.assert_allclose(d, d_expected, rtol=1e-3, atol=1e-4)

def test_alpha_negative():
    """
    Test rowwise concatenation with alpha=-2.0 (scaled second system).

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    n1, m1, p1 = 2, 2, 2
    n2, m2 = 2, 2

    a1 = np.random.randn(n1, n1).astype(float, order='F')
    b1 = np.random.randn(n1, m1).astype(float, order='F')
    c1 = np.random.randn(p1, n1).astype(float, order='F')
    d1 = np.random.randn(p1, m1).astype(float, order='F')

    a2 = np.random.randn(n2, n2).astype(float, order='F')
    b2 = np.random.randn(n2, m2).astype(float, order='F')
    c2 = np.random.randn(p1, n2).astype(float, order='F')
    d2 = np.random.randn(p1, m2).astype(float, order='F')

    alpha = -2.0

    a, b, c, d, n, m, info = ab05od('N', a1, b1, c1, d1, a2, b2, c2, d2, alpha)

    assert info == 0
    assert n == n1 + n2
    assert m == m1 + m2

    a_expected = np.zeros((n, n), order='F', dtype=float)
    a_expected[:n1, :n1] = a1
    a_expected[n1:, n1:] = a2

    b_expected = np.zeros((n, m), order='F', dtype=float)
    b_expected[:n1, :m1] = b1
    b_expected[n1:, m1:] = b2

    c_expected = np.hstack([c1, alpha * c2])
    d_expected = np.hstack([d1, alpha * d2])

    np.testing.assert_allclose(a, a_expected, rtol=1e-14)
    np.testing.assert_allclose(b, b_expected, rtol=1e-14)
    np.testing.assert_allclose(c, c_expected, rtol=1e-14)
    np.testing.assert_allclose(d, d_expected, rtol=1e-14)


"""Test mathematical properties of rowwise concatenation."""

def test_block_diagonal_structure_a():
    """
    Validate A matrix is block diagonal: A = [[A1, 0], [0, A2]].

    The state transition matrix must have zero off-diagonal blocks.
    Random seed: 100 (for reproducibility)
    """
    np.random.seed(100)
    n1, m1, p1 = 3, 2, 2
    n2, m2 = 2, 2

    a1 = np.random.randn(n1, n1).astype(float, order='F')
    b1 = np.random.randn(n1, m1).astype(float, order='F')
    c1 = np.random.randn(p1, n1).astype(float, order='F')
    d1 = np.random.randn(p1, m1).astype(float, order='F')

    a2 = np.random.randn(n2, n2).astype(float, order='F')
    b2 = np.random.randn(n2, m2).astype(float, order='F')
    c2 = np.random.randn(p1, n2).astype(float, order='F')
    d2 = np.random.randn(p1, m2).astype(float, order='F')

    alpha = 1.5

    a, b, c, d, n, m, info = ab05od('N', a1, b1, c1, d1, a2, b2, c2, d2, alpha)

    assert info == 0

    np.testing.assert_allclose(a[:n1, :n1], a1, rtol=1e-14)
    np.testing.assert_allclose(a[n1:, n1:], a2, rtol=1e-14)
    np.testing.assert_allclose(a[:n1, n1:], np.zeros((n1, n2)), rtol=1e-14, atol=1e-15)
    np.testing.assert_allclose(a[n1:, :n1], np.zeros((n2, n1)), rtol=1e-14, atol=1e-15)

def test_block_diagonal_structure_b():
    """
    Validate B matrix is block diagonal: B = [[B1, 0], [0, B2]].

    The input matrix must have zero off-diagonal blocks.
    Random seed: 200 (for reproducibility)
    """
    np.random.seed(200)
    n1, m1, p1 = 3, 2, 2
    n2, m2 = 2, 3

    a1 = np.random.randn(n1, n1).astype(float, order='F')
    b1 = np.random.randn(n1, m1).astype(float, order='F')
    c1 = np.random.randn(p1, n1).astype(float, order='F')
    d1 = np.random.randn(p1, m1).astype(float, order='F')

    a2 = np.random.randn(n2, n2).astype(float, order='F')
    b2 = np.random.randn(n2, m2).astype(float, order='F')
    c2 = np.random.randn(p1, n2).astype(float, order='F')
    d2 = np.random.randn(p1, m2).astype(float, order='F')

    alpha = 0.5

    a, b, c, d, n, m, info = ab05od('N', a1, b1, c1, d1, a2, b2, c2, d2, alpha)

    assert info == 0

    np.testing.assert_allclose(b[:n1, :m1], b1, rtol=1e-14)
    np.testing.assert_allclose(b[n1:, m1:], b2, rtol=1e-14)
    np.testing.assert_allclose(b[:n1, m1:], np.zeros((n1, m2)), rtol=1e-14, atol=1e-15)
    np.testing.assert_allclose(b[n1:, :m1], np.zeros((n2, m1)), rtol=1e-14, atol=1e-15)

def test_output_matrix_concatenation():
    """
    Validate C matrix: C = [C1, alpha*C2] (rowwise concatenation).

    Random seed: 300 (for reproducibility)
    """
    np.random.seed(300)
    n1, m1, p1 = 3, 2, 2
    n2, m2 = 2, 2

    a1 = np.random.randn(n1, n1).astype(float, order='F')
    b1 = np.random.randn(n1, m1).astype(float, order='F')
    c1 = np.random.randn(p1, n1).astype(float, order='F')
    d1 = np.random.randn(p1, m1).astype(float, order='F')

    a2 = np.random.randn(n2, n2).astype(float, order='F')
    b2 = np.random.randn(n2, m2).astype(float, order='F')
    c2 = np.random.randn(p1, n2).astype(float, order='F')
    d2 = np.random.randn(p1, m2).astype(float, order='F')

    for alpha in [1.0, -1.0, 2.5, 0.0]:
        a, b, c, d, n, m, info = ab05od('N', a1, b1, c1, d1, a2, b2, c2, d2, alpha)

        assert info == 0

        np.testing.assert_allclose(c[:, :n1], c1, rtol=1e-14)
        np.testing.assert_allclose(c[:, n1:], alpha * c2, rtol=1e-14)

def test_feedthrough_matrix_concatenation():
    """
    Validate D matrix: D = [D1, alpha*D2] (rowwise concatenation).

    Random seed: 400 (for reproducibility)
    """
    np.random.seed(400)
    n1, m1, p1 = 2, 2, 3
    n2, m2 = 3, 2

    a1 = np.random.randn(n1, n1).astype(float, order='F')
    b1 = np.random.randn(n1, m1).astype(float, order='F')
    c1 = np.random.randn(p1, n1).astype(float, order='F')
    d1 = np.random.randn(p1, m1).astype(float, order='F')

    a2 = np.random.randn(n2, n2).astype(float, order='F')
    b2 = np.random.randn(n2, m2).astype(float, order='F')
    c2 = np.random.randn(p1, n2).astype(float, order='F')
    d2 = np.random.randn(p1, m2).astype(float, order='F')

    for alpha in [1.0, -1.0, 3.14159, 0.0]:
        a, b, c, d, n, m, info = ab05od('N', a1, b1, c1, d1, a2, b2, c2, d2, alpha)

        assert info == 0

        np.testing.assert_allclose(d[:, :m1], d1, rtol=1e-14)
        np.testing.assert_allclose(d[:, m1:], alpha * d2, rtol=1e-14)

def test_state_space_simulation():
    """
    Validate state-space equations for parallel system.

    Test that:
        x(k+1) = A*x(k) + B*u(k)
        y(k) = C*x(k) + D*u(k)

    equals:
        x1(k+1) = A1*x1(k) + B1*u1(k)
        x2(k+1) = A2*x2(k) + B2*u2(k)
        y(k) = C1*x1(k) + D1*u1(k) + alpha*(C2*x2(k) + D2*u2(k))

    Random seed: 500 (for reproducibility)
    """
    np.random.seed(500)
    n1, m1, p1 = 2, 2, 2
    n2, m2 = 2, 2

    a1 = np.random.randn(n1, n1).astype(float, order='F') * 0.5
    b1 = np.random.randn(n1, m1).astype(float, order='F')
    c1 = np.random.randn(p1, n1).astype(float, order='F')
    d1 = np.random.randn(p1, m1).astype(float, order='F')

    a2 = np.random.randn(n2, n2).astype(float, order='F') * 0.5
    b2 = np.random.randn(n2, m2).astype(float, order='F')
    c2 = np.random.randn(p1, n2).astype(float, order='F')
    d2 = np.random.randn(p1, m2).astype(float, order='F')

    alpha = 1.5

    a, b, c, d, n, m, info = ab05od('N', a1, b1, c1, d1, a2, b2, c2, d2, alpha)

    assert info == 0

    x1 = np.array([[1.0], [0.5]], order='F', dtype=float)
    x2 = np.array([[-0.5], [1.0]], order='F', dtype=float)
    x_combined = np.vstack([x1, x2])

    u1 = np.array([[0.5], [-0.3]], order='F', dtype=float)
    u2 = np.array([[0.2], [0.8]], order='F', dtype=float)
    u_combined = np.vstack([u1, u2])

    y_parallel = c1 @ x1 + d1 @ u1 + alpha * (c2 @ x2 + d2 @ u2)
    y_combined = c @ x_combined + d @ u_combined
    np.testing.assert_allclose(y_combined, y_parallel, rtol=1e-14)

    x1_next = a1 @ x1 + b1 @ u1
    x2_next = a2 @ x2 + b2 @ u2
    x_next_parallel = np.vstack([x1_next, x2_next])
    x_next_combined = a @ x_combined + b @ u_combined
    np.testing.assert_allclose(x_next_combined, x_next_parallel, rtol=1e-14)


"""Edge cases and boundary conditions."""

def test_zero_state_first_system():
    """
    Test with N1=0 (first system is static gain D1 only).

    Random seed: 111 (for reproducibility)
    """
    np.random.seed(111)
    m1, p1 = 2, 2
    n2, m2 = 3, 2

    a1 = np.zeros((0, 0), order='F', dtype=float)
    b1 = np.zeros((0, m1), order='F', dtype=float)
    c1 = np.zeros((p1, 0), order='F', dtype=float)
    d1 = np.array([[1.0, 0.5], [0.2, 1.0]], order='F', dtype=float)

    a2 = np.random.randn(n2, n2).astype(float, order='F') * 0.5
    b2 = np.random.randn(n2, m2).astype(float, order='F')
    c2 = np.random.randn(p1, n2).astype(float, order='F')
    d2 = np.random.randn(p1, m2).astype(float, order='F')

    alpha = 2.0

    a, b, c, d, n, m, info = ab05od('N', a1, b1, c1, d1, a2, b2, c2, d2, alpha)

    assert info == 0
    assert n == n2
    assert m == m1 + m2

    np.testing.assert_allclose(a, a2, rtol=1e-14)
    np.testing.assert_allclose(b[:, m1:], b2, rtol=1e-14)
    np.testing.assert_allclose(b[:, :m1], np.zeros((n2, m1)), rtol=1e-14, atol=1e-15)
    np.testing.assert_allclose(c, alpha * c2, rtol=1e-14)
    np.testing.assert_allclose(d[:, :m1], d1, rtol=1e-14)
    np.testing.assert_allclose(d[:, m1:], alpha * d2, rtol=1e-14)

def test_zero_state_second_system():
    """
    Test with N2=0 (second system is static gain D2 only).

    Random seed: 222 (for reproducibility)
    """
    np.random.seed(222)
    n1, m1, p1 = 3, 2, 2
    m2 = 2

    a1 = np.random.randn(n1, n1).astype(float, order='F') * 0.5
    b1 = np.random.randn(n1, m1).astype(float, order='F')
    c1 = np.random.randn(p1, n1).astype(float, order='F')
    d1 = np.random.randn(p1, m1).astype(float, order='F')

    a2 = np.zeros((0, 0), order='F', dtype=float)
    b2 = np.zeros((0, m2), order='F', dtype=float)
    c2 = np.zeros((p1, 0), order='F', dtype=float)
    d2 = np.array([[0.5, 0.2], [0.1, 0.4]], order='F', dtype=float)

    alpha = -1.5

    a, b, c, d, n, m, info = ab05od('N', a1, b1, c1, d1, a2, b2, c2, d2, alpha)

    assert info == 0
    assert n == n1
    assert m == m1 + m2

    np.testing.assert_allclose(a, a1, rtol=1e-14)
    np.testing.assert_allclose(b[:, :m1], b1, rtol=1e-14)
    np.testing.assert_allclose(b[:, m1:], np.zeros((n1, m2)), rtol=1e-14, atol=1e-15)
    np.testing.assert_allclose(c, c1, rtol=1e-14)
    np.testing.assert_allclose(d[:, :m1], d1, rtol=1e-14)
    np.testing.assert_allclose(d[:, m1:], alpha * d2, rtol=1e-14)

def test_both_systems_zero_state():
    """
    Test with N1=0, N2=0 (both systems are static gains only).

    Random seed: 333 (for reproducibility)
    """
    np.random.seed(333)
    m1, m2, p1 = 2, 2, 2

    a1 = np.zeros((0, 0), order='F', dtype=float)
    b1 = np.zeros((0, m1), order='F', dtype=float)
    c1 = np.zeros((p1, 0), order='F', dtype=float)
    d1 = np.random.randn(p1, m1).astype(float, order='F')

    a2 = np.zeros((0, 0), order='F', dtype=float)
    b2 = np.zeros((0, m2), order='F', dtype=float)
    c2 = np.zeros((p1, 0), order='F', dtype=float)
    d2 = np.random.randn(p1, m2).astype(float, order='F')

    alpha = 3.0

    a, b, c, d, n, m, info = ab05od('N', a1, b1, c1, d1, a2, b2, c2, d2, alpha)

    assert info == 0
    assert n == 0
    assert m == m1 + m2

    np.testing.assert_allclose(d[:, :m1], d1, rtol=1e-14)
    np.testing.assert_allclose(d[:, m1:], alpha * d2, rtol=1e-14)

def test_single_state_systems():
    """
    Test with minimal N1=1, N2=1 systems.

    Random seed: 444 (for reproducibility)
    """
    np.random.seed(444)
    n1, m1, p1 = 1, 1, 1
    n2, m2 = 1, 1

    a1 = np.array([[0.5]], order='F', dtype=float)
    b1 = np.array([[1.0]], order='F', dtype=float)
    c1 = np.array([[2.0]], order='F', dtype=float)
    d1 = np.array([[0.1]], order='F', dtype=float)

    a2 = np.array([[0.3]], order='F', dtype=float)
    b2 = np.array([[1.5]], order='F', dtype=float)
    c2 = np.array([[0.8]], order='F', dtype=float)
    d2 = np.array([[0.2]], order='F', dtype=float)

    for alpha in [1.0, -1.0, 0.0, 2.5]:
        a, b, c, d, n, m, info = ab05od('N', a1, b1, c1, d1, a2, b2, c2, d2, alpha)

        assert info == 0
        assert n == 2
        assert m == 2

        np.testing.assert_allclose(a[0, 0], a1[0, 0], rtol=1e-14)
        np.testing.assert_allclose(a[1, 1], a2[0, 0], rtol=1e-14)
        np.testing.assert_allclose(a[0, 1], 0.0, rtol=1e-14, atol=1e-15)
        np.testing.assert_allclose(a[1, 0], 0.0, rtol=1e-14, atol=1e-15)

        np.testing.assert_allclose(c[0, 0], c1[0, 0], rtol=1e-14)
        np.testing.assert_allclose(c[0, 1], alpha * c2[0, 0], rtol=1e-14)

def test_alpha_zero():
    """
    Test with alpha=0.0 (second system contributes nothing to output).

    Random seed: 555 (for reproducibility)
    """
    np.random.seed(555)
    n1, m1, p1 = 2, 2, 2
    n2, m2 = 2, 2

    a1 = np.random.randn(n1, n1).astype(float, order='F') * 0.5
    b1 = np.random.randn(n1, m1).astype(float, order='F')
    c1 = np.random.randn(p1, n1).astype(float, order='F')
    d1 = np.random.randn(p1, m1).astype(float, order='F')

    a2 = np.random.randn(n2, n2).astype(float, order='F') * 0.5
    b2 = np.random.randn(n2, m2).astype(float, order='F')
    c2 = np.random.randn(p1, n2).astype(float, order='F')
    d2 = np.random.randn(p1, m2).astype(float, order='F')

    alpha = 0.0

    a, b, c, d, n, m, info = ab05od('N', a1, b1, c1, d1, a2, b2, c2, d2, alpha)

    assert info == 0

    np.testing.assert_allclose(c[:, n1:], np.zeros((p1, n2)), rtol=1e-14, atol=1e-15)
    np.testing.assert_allclose(d[:, m1:], np.zeros((p1, m2)), rtol=1e-14, atol=1e-15)


"""Test error handling and invalid input detection."""

def test_invalid_over():
    """Test invalid OVER parameter."""
    a1 = np.array([[1.0]], order='F', dtype=float)
    b1 = np.array([[1.0]], order='F', dtype=float)
    c1 = np.array([[1.0]], order='F', dtype=float)
    d1 = np.array([[1.0]], order='F', dtype=float)
    a2 = np.array([[1.0]], order='F', dtype=float)
    b2 = np.array([[1.0]], order='F', dtype=float)
    c2 = np.array([[1.0]], order='F', dtype=float)
    d2 = np.array([[1.0]], order='F', dtype=float)

    with pytest.raises(ValueError, match="over"):
        ab05od('X', a1, b1, c1, d1, a2, b2, c2, d2, 1.0)

def test_dimension_mismatch_p1():
    """Test dimension mismatch: P1 from C1 != P1 from C2."""
    a1 = np.array([[1.0]], order='F', dtype=float)
    b1 = np.array([[1.0]], order='F', dtype=float)
    c1 = np.array([[1.0], [2.0]], order='F', dtype=float)  # P1=2
    d1 = np.array([[1.0], [2.0]], order='F', dtype=float)

    a2 = np.array([[1.0]], order='F', dtype=float)
    b2 = np.array([[1.0]], order='F', dtype=float)
    c2 = np.array([[1.0], [2.0], [3.0]], order='F', dtype=float)  # P1=3 (mismatch)
    d2 = np.array([[1.0]], order='F', dtype=float)

    with pytest.raises(ValueError, match="[pP]1"):
        ab05od('N', a1, b1, c1, d1, a2, b2, c2, d2, 1.0)


"""Test OVER='O' overlap mode."""

def test_overlap_mode_same_results():
    """
    Verify OVER='O' produces same results as OVER='N'.

    The Python wrapper allocates separate output arrays, so the results
    should be identical regardless of OVER mode.

    Random seed: 666 (for reproducibility)
    """
    np.random.seed(666)
    n1, m1, p1 = 2, 2, 2
    n2, m2 = 2, 2

    a1 = np.random.randn(n1, n1).astype(float, order='F') * 0.5
    b1 = np.random.randn(n1, m1).astype(float, order='F')
    c1 = np.random.randn(p1, n1).astype(float, order='F')
    d1 = np.random.randn(p1, m1).astype(float, order='F')

    a2 = np.random.randn(n2, n2).astype(float, order='F') * 0.5
    b2 = np.random.randn(n2, m2).astype(float, order='F')
    c2 = np.random.randn(p1, n2).astype(float, order='F')
    d2 = np.random.randn(p1, m2).astype(float, order='F')

    alpha = -1.0

    a_n, b_n, c_n, d_n, n_n, m_n, info_n = ab05od(
        'N', a1.copy(), b1.copy(), c1.copy(), d1.copy(),
        a2.copy(), b2.copy(), c2.copy(), d2.copy(), alpha)

    a_o, b_o, c_o, d_o, n_o, m_o, info_o = ab05od(
        'O', a1.copy(), b1.copy(), c1.copy(), d1.copy(),
        a2.copy(), b2.copy(), c2.copy(), d2.copy(), alpha)

    assert info_n == 0
    assert info_o == 0
    assert n_n == n_o
    assert m_n == m_o

    np.testing.assert_allclose(a_o, a_n, rtol=1e-14)
    np.testing.assert_allclose(b_o, b_n, rtol=1e-14)
    np.testing.assert_allclose(c_o, c_n, rtol=1e-14)
    np.testing.assert_allclose(d_o, d_n, rtol=1e-14)



"""
Tests for MB05MY: Schur form, eigenvalues, and right eigenvectors.

Computes for an N-by-N real nonsymmetric matrix A:
  - Orthogonal matrix Q reducing A to real Schur form T
  - Eigenvalues (WR + i*WI)
  - Right eigenvectors R of T (upper triangular by construction)
"""
import numpy as np
import pytest
from slicot import mb05my


"""Basic functionality tests."""

def test_small_real_eigenvalues():
    """
    Test 3x3 matrix with real eigenvalues.

    Matrix has eigenvalues 1, 2, 3.
    Random seed: 42 (for reproducibility)
    """
    a = np.array([
        [1.0, 2.0, 3.0],
        [0.0, 2.0, 1.0],
        [0.0, 0.0, 3.0],
    ], order='F', dtype=float)

    wr, wi, r, q, t, info = mb05my('N', a)

    assert info == 0

    eig_computed = np.sort(wr)
    eig_expected = np.array([1.0, 2.0, 3.0])
    np.testing.assert_allclose(eig_computed, eig_expected, rtol=1e-14)

    np.testing.assert_allclose(wi, np.zeros(3), atol=1e-14)

def test_complex_eigenvalues_2x2():
    """
    Test 2x2 rotation matrix with complex conjugate eigenvalues.

    Matrix [[0, -1], [1, 0]] has eigenvalues +-i.
    """
    a = np.array([
        [0.0, -1.0],
        [1.0, 0.0],
    ], order='F', dtype=float)

    wr, wi, r, q, t, info = mb05my('N', a)

    assert info == 0

    np.testing.assert_allclose(wr, [0.0, 0.0], atol=1e-14)
    np.testing.assert_allclose(np.abs(wi), [1.0, 1.0], rtol=1e-14)
    assert wi[0] == -wi[1]

def test_4x4_mixed_eigenvalues():
    """
    Test 4x4 matrix with 2 real and 2 complex eigenvalues.

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    n = 4
    a = np.random.randn(n, n).astype(float, order='F')

    wr, wi, r, q, t, info = mb05my('N', a)

    assert info == 0

    eig_numpy = np.linalg.eigvals(a)
    eig_computed = wr + 1j * wi

    np.testing.assert_allclose(
        sorted(eig_numpy, key=lambda x: (x.real, x.imag)),
        sorted(eig_computed, key=lambda x: (x.real, x.imag)),
        rtol=1e-12
    )


"""Mathematical property tests for Schur decomposition."""

def test_orthogonality_of_q():
    """
    Validate Q is orthogonal: Q^T * Q = I.

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    n = 5
    a = np.random.randn(n, n).astype(float, order='F')

    wr, wi, r, q, t, info = mb05my('N', a)

    assert info == 0
    np.testing.assert_allclose(q.T @ q, np.eye(n), rtol=1e-14, atol=1e-14)
    np.testing.assert_allclose(q @ q.T, np.eye(n), rtol=1e-14, atol=1e-14)

def test_schur_decomposition_a_equals_q_t_qt():
    """
    Validate Schur decomposition: A = Q * T * Q^T.

    Random seed: 456 (for reproducibility)
    """
    np.random.seed(456)
    n = 4
    a = np.random.randn(n, n).astype(float, order='F')
    a_copy = a.copy()

    wr, wi, r, q, t, info = mb05my('N', a)

    assert info == 0
    a_reconstructed = q @ t @ q.T
    np.testing.assert_allclose(a_reconstructed, a_copy, rtol=1e-13, atol=1e-14)

def test_t_is_quasi_upper_triangular():
    """
    Validate T is quasi-upper triangular (real Schur form).

    For real eigenvalues: T is strictly upper triangular at that row.
    For complex pairs: 2x2 block on diagonal.

    Random seed: 789 (for reproducibility)
    """
    np.random.seed(789)
    n = 5
    a = np.random.randn(n, n).astype(float, order='F')

    wr, wi, r, q, t, info = mb05my('N', a)

    assert info == 0

    for i in range(n - 1):
        if wi[i] == 0.0 and wi[i + 1] == 0.0:
            np.testing.assert_allclose(t[i + 1, i], 0.0, atol=1e-14)

def test_eigenvalue_preservation():
    """
    Validate eigenvalues of T match eigenvalues of A.

    Random seed: 888 (for reproducibility)
    """
    np.random.seed(888)
    n = 6
    a = np.random.randn(n, n).astype(float, order='F')

    wr, wi, r, q, t, info = mb05my('N', a)

    assert info == 0

    eig_a = np.linalg.eigvals(a)
    eig_t = np.linalg.eigvals(t)

    np.testing.assert_allclose(
        sorted(eig_a, key=lambda x: (x.real, x.imag)),
        sorted(eig_t, key=lambda x: (x.real, x.imag)),
        rtol=1e-13
    )


"""Mathematical property tests for eigenvectors."""

def test_eigenvector_equation_real():
    """
    Validate T * r = lambda * r for real eigenvalues.

    Random seed: 111 (for reproducibility)
    """
    a = np.array([
        [2.0, 1.0, 0.0],
        [0.0, 3.0, 1.0],
        [0.0, 0.0, 4.0],
    ], order='F', dtype=float)

    wr, wi, r, q, t, info = mb05my('N', a)

    assert info == 0

    for j in range(3):
        if wi[j] == 0.0:
            rj = r[:, j]
            lhs = t @ rj
            rhs = wr[j] * rj
            np.testing.assert_allclose(lhs, rhs, rtol=1e-14, atol=1e-14)

def test_r_is_upper_triangular():
    """
    Validate R is upper triangular by construction.

    Random seed: 222 (for reproducibility)
    """
    np.random.seed(222)
    n = 5
    a = np.random.randn(n, n).astype(float, order='F')

    wr, wi, r, q, t, info = mb05my('N', a)

    assert info == 0

    for j in range(n):
        for i in range(j + 1, n):
            np.testing.assert_allclose(r[i, j], 0.0, atol=1e-14)


"""Tests for diagonal scaling (BALANC='S')."""

def test_with_scaling():
    """
    Test with BALANC='S' for badly scaled matrix.

    Random seed: 333 (for reproducibility)
    """
    np.random.seed(333)
    n = 4
    d = np.diag([1e-5, 1.0, 1e3, 1e6])
    a_balanced = np.random.randn(n, n).astype(float, order='F')
    a = d @ a_balanced @ np.linalg.inv(d)
    a = np.asfortranarray(a)

    wr, wi, r, q, t, info = mb05my('S', a)

    assert info == 0

    eig_numpy = np.linalg.eigvals(a)
    eig_computed = wr + 1j * wi

    np.testing.assert_allclose(
        sorted(eig_numpy, key=lambda x: (x.real, x.imag)),
        sorted(eig_computed, key=lambda x: (x.real, x.imag)),
        rtol=1e-10
    )

def test_no_scaling():
    """
    Test with BALANC='N' (no scaling).

    Random seed: 444 (for reproducibility)
    """
    np.random.seed(444)
    n = 3
    a = np.random.randn(n, n).astype(float, order='F')

    wr, wi, r, q, t, info = mb05my('N', a)

    assert info == 0

    eig_numpy = np.linalg.eigvals(a)
    eig_computed = wr + 1j * wi

    np.testing.assert_allclose(
        sorted(eig_numpy, key=lambda x: (x.real, x.imag)),
        sorted(eig_computed, key=lambda x: (x.real, x.imag)),
        rtol=1e-14
    )


"""Edge case tests."""

def test_n_equals_zero():
    """Test with n=0 (empty matrix)."""
    a = np.array([], order='F', dtype=float).reshape(0, 0)

    wr, wi, r, q, t, info = mb05my('N', a)

    assert info == 0
    assert wr.shape == (0,)
    assert wi.shape == (0,)
    assert r.shape == (0, 0)
    assert q.shape == (0, 0)
    assert t.shape == (0, 0)

def test_n_equals_one():
    """Test scalar case."""
    a = np.array([[3.5]], order='F', dtype=float)

    wr, wi, r, q, t, info = mb05my('N', a)

    assert info == 0
    np.testing.assert_allclose(wr[0], 3.5, rtol=1e-14)
    np.testing.assert_allclose(wi[0], 0.0, atol=1e-14)
    np.testing.assert_allclose(q[0, 0], 1.0, rtol=1e-14)
    np.testing.assert_allclose(t[0, 0], 3.5, rtol=1e-14)

def test_identity_matrix():
    """Test identity matrix (eigenvalues all 1)."""
    n = 4
    a = np.eye(n, order='F', dtype=float)

    wr, wi, r, q, t, info = mb05my('N', a)

    assert info == 0
    np.testing.assert_allclose(wr, np.ones(n), rtol=1e-14)
    np.testing.assert_allclose(wi, np.zeros(n), atol=1e-14)
    np.testing.assert_allclose(t, np.eye(n), rtol=1e-14, atol=1e-14)


"""Error handling tests."""

def test_invalid_balanc():
    """Test that invalid BALANC returns info=-1."""
    a = np.array([[1.0, 2.0], [3.0, 4.0]], order='F', dtype=float)

    wr, wi, r, q, t, info = mb05my('X', a)

    assert info == -1



"""
Tests for SB16AY - Cholesky factors of frequency-weighted controllability
and observability Grammians for controller reduction.

SB16AY computes for given state-space representations (A,B,C,D) and
(Ac,Bc,Cc,Dc) of the open-loop system G and feedback controller K,
the Cholesky factors S and R of the frequency-weighted Grammians
P = S*S' (controllability) and Q = R'*R (observability).

The controller must stabilize the closed-loop system.
Ac must be in block-diagonal real Schur form: Ac = diag(Ac1, Ac2),
where Ac1 contains unstable eigenvalues and Ac2 contains stable eigenvalues.
"""
import ctypes
import numpy as np
import os
import glob
import pytest


def find_slicot_library():
    """Find the slicot shared library."""
    project_root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    build_dirs = [
        os.path.join(project_root, 'build', 'macos-arm64-debug', 'src'),
        os.path.join(project_root, 'build', 'linux-x64-debug', 'src'),
        os.path.join(project_root, 'build', 'linux-x64-debug-sanitizers', 'src'),
        os.path.join(project_root, 'build', 'macos-arm64-release', 'src'),
        os.path.join(project_root, 'build', 'linux-x64-release', 'src'),
    ]
    for bd in build_dirs:
        if os.path.exists(bd):
            libs = glob.glob(os.path.join(bd, 'libslicot.*'))
            if libs:
                return libs[0]
    pytest.skip("Could not find libslicot shared library")


@pytest.fixture(scope='module')
def lib():
    """Load the slicot library."""
    lib_path = find_slicot_library()
    slicot = ctypes.CDLL(lib_path)
    return slicot


@pytest.fixture(scope='module')
def sb16ay(lib):
    """Get the sb16ay function with proper signature."""
    func = lib.sb16ay
    func.argtypes = [
        ctypes.c_char_p,                  # dico
        ctypes.c_char_p,                  # jobc
        ctypes.c_char_p,                  # jobo
        ctypes.c_char_p,                  # weight
        ctypes.c_int,                     # n
        ctypes.c_int,                     # m
        ctypes.c_int,                     # p
        ctypes.c_int,                     # nc
        ctypes.c_int,                     # ncs
        ctypes.POINTER(ctypes.c_double),  # a
        ctypes.c_int,                     # lda
        ctypes.POINTER(ctypes.c_double),  # b
        ctypes.c_int,                     # ldb
        ctypes.POINTER(ctypes.c_double),  # c
        ctypes.c_int,                     # ldc
        ctypes.POINTER(ctypes.c_double),  # d
        ctypes.c_int,                     # ldd
        ctypes.POINTER(ctypes.c_double),  # ac
        ctypes.c_int,                     # ldac
        ctypes.POINTER(ctypes.c_double),  # bc
        ctypes.c_int,                     # ldbc
        ctypes.POINTER(ctypes.c_double),  # cc
        ctypes.c_int,                     # ldcc
        ctypes.POINTER(ctypes.c_double),  # dc
        ctypes.c_int,                     # lddc
        ctypes.POINTER(ctypes.c_double),  # scalec
        ctypes.POINTER(ctypes.c_double),  # scaleo
        ctypes.POINTER(ctypes.c_double),  # s
        ctypes.c_int,                     # lds
        ctypes.POINTER(ctypes.c_double),  # r
        ctypes.c_int,                     # ldr
        ctypes.POINTER(ctypes.c_int),     # iwork
        ctypes.POINTER(ctypes.c_double),  # dwork
        ctypes.c_int,                     # ldwork
        ctypes.POINTER(ctypes.c_int),     # info
    ]
    func.restype = None
    return func


def test_no_weighting_continuous(sb16ay):
    """
    Test WEIGHT='N' for continuous-time.

    When no weighting is used:
    - Controllability Grammian P solves: Ac2*P + P*Ac2' + scalec^2*Bc2*Bc2' = 0
    - Observability Grammian Q solves: Ac2'*Q + Q*Ac2 + scaleo^2*Cc2'*Cc2 = 0

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)

    n, m, p = 2, 1, 1
    nc, ncs = 3, 3

    a = np.array([[-1.0, 0.5], [0.0, -2.0]], order='F', dtype=np.float64)
    b = np.array([[1.0], [0.0]], order='F', dtype=np.float64)
    c = np.array([[1.0, 1.0]], order='F', dtype=np.float64)
    d = np.array([[0.0]], order='F', dtype=np.float64)

    ac = np.array([
        [-0.5, 0.1, 0.0],
        [0.0, -1.0, 0.2],
        [0.0, 0.0, -1.5]
    ], order='F', dtype=np.float64)
    bc = np.array([[1.0], [0.5], [0.0]], order='F', dtype=np.float64)
    cc = np.array([[0.5, 0.3, 0.2]], order='F', dtype=np.float64)
    dc = np.array([[0.1]], order='F', dtype=np.float64)

    s = np.zeros((ncs, ncs), order='F', dtype=np.float64)
    r = np.zeros((ncs, ncs), order='F', dtype=np.float64)
    scalec = ctypes.c_double(0.0)
    scaleo = ctypes.c_double(0.0)

    ldwork = ncs * (max(m, p) + 5)
    ldwork = max(1, ldwork)
    dwork = np.zeros(ldwork, dtype=np.float64)
    iwork = np.zeros(1, dtype=np.int32)
    info = ctypes.c_int(0)

    sb16ay(
        b"C", b"S", b"S", b"N",
        n, m, p, nc, ncs,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        c.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), p,
        d.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), p,
        ac.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), nc,
        bc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), nc,
        cc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), m,
        dc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), m,
        ctypes.byref(scalec), ctypes.byref(scaleo),
        s.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ncs,
        r.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ncs,
        iwork.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldwork,
        ctypes.byref(info)
    )

    assert info.value == 0
    assert scalec.value > 0
    assert scaleo.value > 0

    s_upper = np.triu(s)
    r_upper = np.triu(r)
    p = s_upper @ s_upper.T
    q = r_upper.T @ r_upper

    bc2 = bc
    cc2 = cc

    res_c = ac @ p + p @ ac.T + scalec.value**2 * bc2 @ bc2.T
    res_o = ac.T @ q + q @ ac + scaleo.value**2 * cc2.T @ cc2

    np.testing.assert_allclose(res_c, 0.0, atol=1e-10)
    np.testing.assert_allclose(res_o, 0.0, atol=1e-10)


def test_no_weighting_discrete(sb16ay):
    """
    Test WEIGHT='N' for discrete-time.

    When no weighting is used:
    - Controllability Grammian P solves: Ac2*P*Ac2' - P + scalec^2*Bc2*Bc2' = 0
    - Observability Grammian Q solves: Ac2'*Q*Ac2 - Q + scaleo^2*Cc2'*Cc2 = 0

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)

    n, m, p = 2, 1, 1
    nc, ncs = 3, 3

    a = np.array([[0.5, 0.1], [0.0, 0.4]], order='F', dtype=np.float64)
    b = np.array([[1.0], [0.0]], order='F', dtype=np.float64)
    c = np.array([[1.0, 1.0]], order='F', dtype=np.float64)
    d = np.array([[0.0]], order='F', dtype=np.float64)

    ac = np.array([
        [0.3, 0.1, 0.0],
        [0.0, 0.4, 0.1],
        [0.0, 0.0, 0.2]
    ], order='F', dtype=np.float64)
    bc = np.array([[1.0], [0.5], [0.0]], order='F', dtype=np.float64)
    cc = np.array([[0.5, 0.3, 0.2]], order='F', dtype=np.float64)
    dc = np.array([[0.1]], order='F', dtype=np.float64)

    s = np.zeros((ncs, ncs), order='F', dtype=np.float64)
    r = np.zeros((ncs, ncs), order='F', dtype=np.float64)
    scalec = ctypes.c_double(0.0)
    scaleo = ctypes.c_double(0.0)

    ldwork = ncs * (max(m, p) + 5)
    ldwork = max(1, ldwork)
    dwork = np.zeros(ldwork, dtype=np.float64)
    iwork = np.zeros(1, dtype=np.int32)
    info = ctypes.c_int(0)

    sb16ay(
        b"D", b"S", b"S", b"N",
        n, m, p, nc, ncs,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        c.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), p,
        d.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), p,
        ac.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), nc,
        bc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), nc,
        cc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), m,
        dc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), m,
        ctypes.byref(scalec), ctypes.byref(scaleo),
        s.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ncs,
        r.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ncs,
        iwork.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldwork,
        ctypes.byref(info)
    )

    assert info.value == 0
    assert scalec.value > 0
    assert scaleo.value > 0

    s_upper = np.triu(s)
    r_upper = np.triu(r)
    p = s_upper @ s_upper.T
    q = r_upper.T @ r_upper

    bc2 = bc
    cc2 = cc

    res_c = ac @ p @ ac.T - p + scalec.value**2 * bc2 @ bc2.T
    res_o = ac.T @ q @ ac - q + scaleo.value**2 * cc2.T @ cc2

    np.testing.assert_allclose(res_c, 0.0, atol=1e-10)
    np.testing.assert_allclose(res_o, 0.0, atol=1e-10)


def test_zero_ncs(sb16ay):
    """
    Test with NCS=0 (quick return).
    """
    n, m, p = 2, 1, 1
    nc, ncs = 2, 0

    a = np.array([[-1.0, 0.0], [0.0, -2.0]], order='F', dtype=np.float64)
    b = np.array([[1.0], [0.0]], order='F', dtype=np.float64)
    c = np.array([[1.0, 1.0]], order='F', dtype=np.float64)
    d = np.array([[0.0]], order='F', dtype=np.float64)

    ac = np.array([[1.5, 0.1], [0.0, 2.0]], order='F', dtype=np.float64)
    bc = np.array([[1.0], [0.0]], order='F', dtype=np.float64)
    cc = np.array([[1.0, 0.5]], order='F', dtype=np.float64)
    dc = np.array([[0.1]], order='F', dtype=np.float64)

    s = np.zeros((1, 1), order='F', dtype=np.float64)
    r = np.zeros((1, 1), order='F', dtype=np.float64)
    scalec = ctypes.c_double(0.0)
    scaleo = ctypes.c_double(0.0)

    ldwork = 1
    dwork = np.zeros(ldwork, dtype=np.float64)
    iwork = np.zeros(1, dtype=np.int32)
    info = ctypes.c_int(0)

    sb16ay(
        b"C", b"S", b"S", b"N",
        n, m, p, nc, ncs,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        c.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), p,
        d.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), p,
        ac.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), nc,
        bc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), nc,
        cc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), m,
        dc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), m,
        ctypes.byref(scalec), ctypes.byref(scaleo),
        s.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), 1,
        r.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), 1,
        iwork.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldwork,
        ctypes.byref(info)
    )

    assert info.value == 0
    assert scalec.value == pytest.approx(1.0)
    assert scaleo.value == pytest.approx(1.0)


def test_invalid_dico(sb16ay):
    """Test invalid DICO parameter returns info=-1."""
    n, m, p = 2, 1, 1
    nc, ncs = 2, 2

    a = np.array([[-1.0, 0.0], [0.0, -2.0]], order='F', dtype=np.float64)
    b = np.array([[1.0], [0.0]], order='F', dtype=np.float64)
    c = np.array([[1.0, 1.0]], order='F', dtype=np.float64)
    d = np.array([[0.0]], order='F', dtype=np.float64)
    ac = np.array([[-0.5, 0.0], [0.0, -1.0]], order='F', dtype=np.float64)
    bc = np.array([[1.0], [0.0]], order='F', dtype=np.float64)
    cc = np.array([[1.0, 0.5]], order='F', dtype=np.float64)
    dc = np.array([[0.1]], order='F', dtype=np.float64)

    s = np.zeros((ncs, ncs), order='F', dtype=np.float64)
    r = np.zeros((ncs, ncs), order='F', dtype=np.float64)
    scalec = ctypes.c_double(0.0)
    scaleo = ctypes.c_double(0.0)

    ldwork = ncs * (max(m, p) + 5)
    dwork = np.zeros(ldwork, dtype=np.float64)
    iwork = np.zeros(1, dtype=np.int32)
    info = ctypes.c_int(0)

    sb16ay(
        b"X", b"S", b"S", b"N",
        n, m, p, nc, ncs,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        c.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), p,
        d.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), p,
        ac.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), nc,
        bc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), nc,
        cc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), m,
        dc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), m,
        ctypes.byref(scalec), ctypes.byref(scaleo),
        s.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ncs,
        r.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ncs,
        iwork.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldwork,
        ctypes.byref(info)
    )

    assert info.value == -1


def test_invalid_ncs(sb16ay):
    """Test NCS > NC returns info=-9."""
    n, m, p = 2, 1, 1
    nc, ncs = 2, 3

    a = np.array([[-1.0, 0.0], [0.0, -2.0]], order='F', dtype=np.float64)
    b = np.array([[1.0], [0.0]], order='F', dtype=np.float64)
    c = np.array([[1.0, 1.0]], order='F', dtype=np.float64)
    d = np.array([[0.0]], order='F', dtype=np.float64)
    ac = np.array([[-0.5, 0.0], [0.0, -1.0]], order='F', dtype=np.float64)
    bc = np.array([[1.0], [0.0]], order='F', dtype=np.float64)
    cc = np.array([[1.0, 0.5]], order='F', dtype=np.float64)
    dc = np.array([[0.1]], order='F', dtype=np.float64)

    s = np.zeros((ncs, ncs), order='F', dtype=np.float64)
    r = np.zeros((ncs, ncs), order='F', dtype=np.float64)
    scalec = ctypes.c_double(0.0)
    scaleo = ctypes.c_double(0.0)

    ldwork = ncs * (max(m, p) + 5)
    dwork = np.zeros(ldwork, dtype=np.float64)
    iwork = np.zeros(1, dtype=np.int32)
    info = ctypes.c_int(0)

    sb16ay(
        b"C", b"S", b"S", b"N",
        n, m, p, nc, ncs,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        c.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), p,
        d.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), p,
        ac.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), nc,
        bc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), nc,
        cc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), m,
        dc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), m,
        ctypes.byref(scalec), ctypes.byref(scaleo),
        s.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ncs,
        r.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ncs,
        iwork.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldwork,
        ctypes.byref(info)
    )

    assert info.value == -9


def test_unstable_ac2_continuous(sb16ay):
    """
    Test that unstable Ac2 (stable part) returns info=5 for continuous.
    """
    n, m, p = 2, 1, 1
    nc, ncs = 2, 2

    a = np.array([[-1.0, 0.0], [0.0, -2.0]], order='F', dtype=np.float64)
    b = np.array([[1.0], [0.0]], order='F', dtype=np.float64)
    c = np.array([[1.0, 1.0]], order='F', dtype=np.float64)
    d = np.array([[0.0]], order='F', dtype=np.float64)

    ac = np.array([[1.0, 0.0], [0.0, 2.0]], order='F', dtype=np.float64)
    bc = np.array([[1.0], [0.0]], order='F', dtype=np.float64)
    cc = np.array([[1.0, 0.5]], order='F', dtype=np.float64)
    dc = np.array([[0.1]], order='F', dtype=np.float64)

    s = np.zeros((ncs, ncs), order='F', dtype=np.float64)
    r = np.zeros((ncs, ncs), order='F', dtype=np.float64)
    scalec = ctypes.c_double(0.0)
    scaleo = ctypes.c_double(0.0)

    ldwork = ncs * (max(m, p) + 5)
    dwork = np.zeros(ldwork, dtype=np.float64)
    iwork = np.zeros(1, dtype=np.int32)
    info = ctypes.c_int(0)

    sb16ay(
        b"C", b"S", b"S", b"N",
        n, m, p, nc, ncs,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        c.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), p,
        d.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), p,
        ac.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), nc,
        bc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), nc,
        cc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), m,
        dc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), m,
        ctypes.byref(scalec), ctypes.byref(scaleo),
        s.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ncs,
        r.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ncs,
        iwork.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldwork,
        ctypes.byref(info)
    )

    assert info.value == 5


def test_cholesky_symmetry_continuous(sb16ay):
    """
    Test that S*S' is symmetric positive semi-definite (controllability)
    and R'*R is symmetric positive semi-definite (observability).

    Random seed: 456 (for reproducibility)
    """
    np.random.seed(456)

    n, m, p = 2, 1, 1
    nc, ncs = 4, 4

    a = np.array([[-1.0, 0.2], [-0.1, -2.0]], order='F', dtype=np.float64)
    b = np.array([[1.0], [0.5]], order='F', dtype=np.float64)
    c = np.array([[1.0, 0.5]], order='F', dtype=np.float64)
    d = np.array([[0.0]], order='F', dtype=np.float64)

    ac = np.array([
        [-0.5, 0.1, 0.0, 0.0],
        [0.0, -0.8, 0.1, 0.0],
        [0.0, 0.0, -1.2, 0.1],
        [0.0, 0.0, 0.0, -1.5]
    ], order='F', dtype=np.float64)
    bc = np.array([[1.0], [0.5], [0.3], [0.1]], order='F', dtype=np.float64)
    cc = np.array([[0.5, 0.4, 0.3, 0.2]], order='F', dtype=np.float64)
    dc = np.array([[0.05]], order='F', dtype=np.float64)

    s = np.zeros((ncs, ncs), order='F', dtype=np.float64)
    r = np.zeros((ncs, ncs), order='F', dtype=np.float64)
    scalec = ctypes.c_double(0.0)
    scaleo = ctypes.c_double(0.0)

    ldwork = ncs * (max(m, p) + 5)
    dwork = np.zeros(ldwork, dtype=np.float64)
    iwork = np.zeros(1, dtype=np.int32)
    info = ctypes.c_int(0)

    sb16ay(
        b"C", b"S", b"S", b"N",
        n, m, p, nc, ncs,
        a.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        b.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), n,
        c.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), p,
        d.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), p,
        ac.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), nc,
        bc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), nc,
        cc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), m,
        dc.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), m,
        ctypes.byref(scalec), ctypes.byref(scaleo),
        s.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ncs,
        r.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ncs,
        iwork.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),
        dwork.ctypes.data_as(ctypes.POINTER(ctypes.c_double)), ldwork,
        ctypes.byref(info)
    )

    assert info.value == 0

    s_upper = np.triu(s)
    r_upper = np.triu(r)
    p = s_upper @ s_upper.T
    q = r_upper.T @ r_upper

    np.testing.assert_allclose(p, p.T, rtol=1e-14)
    np.testing.assert_allclose(q, q.T, rtol=1e-14)

    eig_p = np.linalg.eigvalsh(p)
    eig_q = np.linalg.eigvalsh(q)
    assert np.all(eig_p >= -1e-14)
    assert np.all(eig_q >= -1e-14)
